""" infinihost v{version}
INFINIDAT Host PowerTools

Usage:
    infinihost settings check [(--fix | --auto-fix)] [--with-qr-codes]
    infinihost system list
    infinihost system register [(--fix | --auto-fix)] [--hostname=HOSTNAME] [--system=SYSTEM]
    infinihost connectivity list [--json]
    infinihost pool list [--units=UNITS] [--rescan] [--system=SYSTEM]
    infinihost volume list [(--summary | --detailed | --json)] [--units=UNITS] [--rescan]
    infinihost volume provision <size> [--yes] [--name=NAME] [--filesystem=FS] [--mount=PATH] [--thin=THIN] [--pool=POOL] [--system=SYSTEM] [--lun=LUN] [--no-filesystem] [--no-mount]
    infinihost volume resize <device-or-mount> <size> [--expand-by] [--yes]
    infinihost volume create-filesystem <device> [--filesystem=FS]
    infinihost volume mount <device> [<mount-path>] [--allow-read-only]
    infinihost volume unmount <device-or-mount> [--yes]
    infinihost volume unmap <device-or-mount> [--yes]
    infinihost volume map <volume> [--system=SYSTEM] [--yes]
    infinihost volume rename <device-or-mount> <name>
    infinihost volume speedtest <device-or-mount> [--duration=TIME]
    infinihost metadata sync
    infinihost rescan
    infinihost logs collect [(--interactive | --only-infinihost-logs)] [--file=FILE] [--timestamp=DATETIME] [--span=TIME]
    infinihost credentials set [--system=SYSTEM] [<username> [<password>]]
    infinihost credentials get
    infinihost credentials delete

Options:
    <device-or-mount>           device to work with - accepts device path, name, or mount point
    <device>                    device to work with - accepts device path
    <mount-path>                new path for the filesystem mount [default: autodetect]
    <size>                      desired volume size (examples: 10gb, 100mb, 1tb)
    <volume>                    volume id or name to map
    --auto-fix                  automatically attempt to fix all issues
    --detailed                  a much more detailed view
    --duration=TIME             time (in seconds) to run [default: 60]
    --expand-by                 interpret the given size as an addition to the current volume size
    --file=FILE                 output filename or path (optional)
    --filesystem=FS             filesystem to create on new volume [default: autodetect]
    --fix                       ask for confirmation on each fix that can be applied
    --hostname=HOSTNAME         hostname to use when registering [default: autodetect]
    --interactive               interactively choose what to collect
    --json                      JSON-formatted output, useful for scripting
    --lun=LUN                   specific LUN to assign to volume
    --mount=PATH                mount path [default: autodetect]
    --name=NAME                 name of the volume to create [default: autodetect]
    --no-filesystem             do not create a filesystem on the new volume. This also implies --no-mount
    --no-mount                  do not mount the new filesystem
    --only-infinihost-logs      only collect infinihost log files
    --pool=POOL                 pool to provision from, see infinihost system list for the available options [default: autodetect]
    --rescan                    rescan the storage stack before listing
    --span=TIME                 number of seconds to go back from --timestamp to start collecting the logs [default: 86400]
    --summary                   a very short summary
    --system=SYSTEM             system serial or name to use, see infinihost system list for the available options [default: autodetect]
    --thin=THIN                 volume thin provisioning (options: yes, no) [default: no]
    --timestamp=DATETIME        date and time of logs to collect. see --span [default: now]
    --units=UNITS               size unit to use in 'brief' output [default: GB]
    --with-qr-codes             display QR codes in addition to links
    --yes                       provide to confirm the action
    -h --help                   show this screen.
    -v --version                show version.

{privileges_text}{autocompletion_text}

More information:
    infinihost settings check          validate configuration of the storage stack
    infinihost connectivity list       show fiberchannel connectivity
    infinihost pool list               show attached infinibox pools
    infinihost volume list             show attached infinibox volumes; does a rescan beforehand by default
    infinihost volume provision        provision a new volume from infinibox
    infinihost volume resize           change the size of an existing volume
    infinihost volume unmap            unmap a volume from infinibox
    infinihost volume speedtest        run a non-distruptive speed test on a specific volume
    infinihost system list             show information on connected systems
    infinihost system register         register host in system
    infinihost rescan                  rescan the storage stack
    infinihost logs collect            collect diagnostic information
    infinihost metadata sync           updates metadata of host and volumes on the infinibox machine
    infinihost credentials set         save credentials for a specific infinibox machine or globally
    infinihost credentials get         show all saved credentials
    infinihost credentials delete      remove all saved credentials

Getting started:

# Check that this host is ready:
{prefix}infinihost settings check

# Make the host ready:
{prefix}infinihost settings check --auto-fix

# Register the host in the system:
{prefix}infinihost system register

# Provision a new volume, a 10GB in this example:
{prefix}infinihost volume provision 10gb

# You can see it by running:
{prefix}infinihost volume list
"""

import os
import sys
import docopt
from infi.vendata.powertools.__version__ import __version__

def get_privileges_text():
    import colorama
    privileges = "administrative" if os.name == 'nt' else 'root'
    return colorama.Fore.RED + "This tool requires {privileges} privileges.".format(privileges=privileges) + colorama.Fore.RESET

def get_autocompletion_text():
    autocompletion_paths = ["/etc/bash_completion.d/_infinihost.sh",
                            os.path.expanduser("~/.zprezto/modules/completion/external/src/_infinihost"),
                            os.path.expanduser("~/.oh-my-zsh/completions/_infinihost")]
    if any(os.path.exists(path) for path in autocompletion_paths):
        return "\nShell tab auto-completion is available for this tool."
    return ""

def shorten_docopt_exit_string(e):
    # shorten the usage lines printed by DocoptExit to include only lines that are relevant to the command
    # get all the command words in argv (ignore flags and the executable)
    argv = [word for word in sys.argv[1:] if not word.startswith('-') and word in __doc__]
    argv_len = len(argv)
    if argv_len == 0:
        return
    short_usage_lines = [line for line in e.usage.split("\n")
                         if line.split()[1:argv_len+1] == argv]
    if len(short_usage_lines) > 0:
        # 'code' is inherited from SystemExit and it's what's printed to the screen
        # it may contain a message, and it contains the usage that we want to replace
        e.code = e.code[:e.code.find("Usage:\n")] + "Usage:\n" + "\n".join(short_usage_lines)

def parse_commandline_arguments(argv):
    from infi.os_info import get_platform_string
    from . import openstack, veritas, solaris
    if openstack.is_openstack_cinder_installed():
        doc = openstack.__doc__
    elif veritas.is_veritas_installed():
        doc = veritas.__doc__
    elif "solaris" in get_platform_string():
        doc = solaris.__doc__
    elif "aix" in get_platform_string():
        doc = solaris.__doc__.replace("a Solaris", "an AIX")
    else:
        doc = __doc__
    try:
        return docopt.docopt(doc.format(version=__version__,
                                        privileges_text=get_privileges_text(),
                                        autocompletion_text=get_autocompletion_text(),
                                        prefix="" if os.name == 'nt' else 'sudo ').strip(),
                                        version=__version__, help=True, argv=argv)
    except docopt.DocoptExit as e:
        shorten_docopt_exit_string(e)
        raise


def in_cli(argv=sys.argv[1:]):
    arguments = parse_commandline_arguments(argv)
    # we import the engine script here because it contains many imports and may take time to load,
    # but we want the docopt check to run first, as fast as possible
    from .engine import run_command
    from infi.vendata.powertools.utils.broken_pipe import silence_broken_pipe_on_stdout_and_stderr
    import warnings
    with warnings.catch_warnings():
        import infinisdk  # noqa
    silence_broken_pipe_on_stdout_and_stderr()
    return run_command(arguments)

def install_shell_completion():
    from infi.docopt_completion.docopt_completion import docopt_completion
    return docopt_completion("infinihost")

def post_install():
    try:
        install_shell_completion()
    finally:
        return 0

def pre_uninstall():
    from os import path
    from shutil import rmtree
    from infi.vendata.powertools import PROJECTROOT
    try:
        conf_dir = path.join(PROJECTROOT, 'conf')
        if path.exists(conf_dir):
            rmtree(conf_dir, ignore_errors=True)
    finally:
        return 0
