from infi.storagemodel.vendor.infinidat.infinibox.wwn import InfinidatWWN, InvalidInfinidatWWN, InvalidWWN
from logging import getLogger

logger = getLogger(__name__)
NA = "N/A"

def _fix_wwn_string(item):
    from infi.dtypes.wwn import WWN
    if isinstance(item, WWN):
        return str(item)
    return item

def _get_output():
    from infi.hbaapi import get_ports_collection, FC_PORT_ATTRIBUTES, FC_PORT_STATISTICS
    ports = get_ports_collection().get_ports()
    logger.info("ports retrieved from hbaapi: {}".format(ports))
    INTERESTING_ATTRIBUTES = set(FC_PORT_ATTRIBUTES) - set(["discovered_ports", "model", "port_wwn", "port_state"])
    initiator_ports = [{"port_model": initiator_port.model or initiator_port.model_description or initiator_port.manufacturer,
                        "port_name": repr(initiator_port.port_wwn),
                        "port_state": initiator_port.port_state,
                        "remote_ports": [{"port_name": repr(target_port.port_wwn),
                                          "port_state": target_port.port_state,
                                          "attributes": {key: _fix_wwn_string(target_port.get(key))
                                                         for key in INTERESTING_ATTRIBUTES},
                                          "statistics": {key: _fix_wwn_string(target_port.get(key))
                                                         for key in FC_PORT_STATISTICS},
                                         }
                                         for target_port in initiator_port.discovered_ports],
                        "attributes": {key: _fix_wwn_string(initiator_port.get(key)) for key in INTERESTING_ATTRIBUTES},
                        "statistics": {key: _fix_wwn_string(initiator_port.get(key)) for key in FC_PORT_STATISTICS},
                       } for initiator_port in ports]
    return initiator_ports

def print_line(values, lengths):
    line = ''.join([name.ljust(size) for name, size in zip(values, lengths)])
    print line

def handle_remote_port_value(remote_port_value):
    from munch import Munch
    if remote_port_value == NA:
        return [Munch(port_name=NA)]
    return remote_port_value

def get_system_serial(port_wwn):
    try:
        return str(InfinidatWWN(port_wwn).get_system_serial())
    except (InvalidWWN, InvalidInfinidatWWN):
        return NA

def get_system_name(system_serial):
    if system_serial == NA:
        return NA

    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0

    try:
        for device in get_infinidat_block_devices_and_controllers__mapped_to_lun0():
            if str(device.get_vendor().get_system_serial()) == str(system_serial):
                return device.get_vendor().get_system_name()
    except:
        logger.exception("failed to deduce system name for serial {}".format(system_serial))
    return NA

def print_table(ports):
    COLUMN_NAMES = ["Port Model", "State", "Initiator WWPN", "Target WWPN", "InfiniBox"]
    COLUMN_KEYS = ["port_model", "port_state", "port_name", "remote_ports"]
    COLUMN_LENGTHS = [12, 10, 18, 18, 22]

    # print table header
    print_line(COLUMN_NAMES, COLUMN_LENGTHS)
    local_wwpns = [p['port_name'] for p in ports]
    for port in ports:
        values = [(port[key] or NA)[:COLUMN_LENGTHS[index]-2] or NA for index, key in enumerate(COLUMN_KEYS)]
        additional_remote_ports = handle_remote_port_value(values[-1])
        for additional_remote_port in additional_remote_ports:
            if additional_remote_port['port_name'] in local_wwpns:
                continue
            connection = values[:-1] + [additional_remote_port["port_name"]]
            system_serial = get_system_serial(additional_remote_port["port_name"])
            system_name = get_system_name(system_serial)
            connection += [system_name[:22] if system_name != NA else
                           '%s (serial)' % system_serial if system_serial != NA else
                           NA]
            print_line(connection, COLUMN_LENGTHS)

def print_output(json=False):
    from json import dumps
    output = _get_output()
    if json:
        print dumps(output, indent=True)
        return

    if len(output) == 0:
        print "No initiator ports detected on host."
        return

    print_table(output)
