import cherrypy
import logging
import cjson
from munch import Munch
from .....webapp import json_response
from ...resource import ViewBase

logger = logging.getLogger(__name__)

from .auth import Auth
from .readiness import Readiness
from .volumes import Volumes, HostVolumesTable
from .registration import HostRegistration
from .operations import Operations

def get_infiniboxes():
    infiniboxes = []
    from infi.vendata.powertools.utils.queries import get_uniq_list_of_attached_boxes__by_serial
    from infi.vendata.powertools.utils.versioncheck import is_supported
    from infi.vendata.powertools.utils import auth
    box_controllers = get_uniq_list_of_attached_boxes__by_serial()
    for box in box_controllers:
        vendor = box.get_vendor()
        serial = vendor.get_system_serial()
        credentials = auth.CREDENTIALS_STORE.get_existing_credentials(serial)
        username = None if credentials is None else credentials.get_username()
        url = "https://{}".format(vendor.get_management_address())
        infiniboxes.append(Munch(name=vendor.get_system_name(),
                                 serial=serial,
                                 gui=url,
                                 username=username,
                                 version=vendor.get_system_version(),
                                 supported=is_supported(vendor.get_system_version())))
    return infiniboxes

class Host(ViewBase):
    @cherrypy.expose
    def index(self, **kwargs):
        import platform
        from time import time
        from infi.vendata.powertools.cli.openstack import is_openstack_cinder_installed
        from infi.vendata.powertools.utils import is_veritas_multipathing_installed
        from infi.vendata.powertools.__version__ import __version__
        from infi.storagemodel import get_storage_model
        get_storage_model().refresh()  # clear cache so that reloading the page will actually refresh its contents
        url = self.get_url_from_request()
        sources = self._get_sources(url)
        hostname = self._get_hostname()
        timestamp = time()
        is_windows = platform.system() == 'Windows'
        kwargs = dict(sources=cjson.encode(sources),
                      hostname=hostname,
                      timestamp=timestamp,
                      version=__version__,
                      is_openstack_cinder_installed=is_openstack_cinder_installed(),
                      is_windows=is_windows,
                      is_veritas=is_veritas_multipathing_installed(),
                      auth_key=kwargs.get('auth_key'))
        return self.template_lookup.get_template('host.mako').render(**kwargs)

    @json_response
    def get_infinibox_table(self, **kwargs):
        from infi.vendata.powertools.__version__ import __version__
        columns = self._get_columns()
        infiniboxes = self._get_infiniboxes()
        kwargs = dict(infiniboxes=infiniboxes, table_columns=columns, powertools_version=__version__)
        return self.template_lookup.get_template('box_table.mako').render(**kwargs)

    @cherrypy.expose
    def collect_logs(self, **kwargs):
        from cherrypy.lib.static import serve_file
        from infi.vendata.powertools.logs_collect.engine import run
        rc, filepath = run('now', '600', silent=True)
        return serve_file(filepath, "text/html", "attachment")

    def _get_hostname(self):
        from infi.vendata.powertools.utils.queries import get_os_hostname
        return get_os_hostname()

    def _get_infiniboxes(self):
        return get_infiniboxes()

    def _get_sources(self, url):
        url_format = url.replace("/host/?", "/host/{uri}/?")
        sources = Munch(
            status=url_format.format(uri="readiness"),
            registration=url_format.format(uri="registration"),
            volume_table=url_format.format(uri="volumes"),
            inventory=url_format.format(uri="get_infinibox_table"),
            box_volumes=url_format.format(uri="volumes/box"),
            logscollect=url_format.format(uri="collect_logs"),
        )
        sources.auth = Munch(
            has_credentials=url_format.format(uri="auth/has_credentials"),
            set_credentials=url_format.format(uri="auth/set_credentials"),
            reset_credentials=url_format.format(uri="auth/reset_credentials"),
            get_current_username=url_format.format(uri="auth/get_current_username"),
        )
        sources.operations = Munch(
            unmap=url_format.format(uri="operations/unmap"),
            is_read_only=url_format.format(uri="operations/is_read_only"),
            mount=url_format.format(uri="operations/mount"),
            unmount=url_format.format(uri="operations/unmount"),
            rename=url_format.format(uri="operations/rename"),
            rescan=url_format.format(uri="operations/rescan"),
            delete=url_format.format(uri="operations/delete"),
            create_fs=url_format.format(uri="operations/create_fs"),
            run_provision_task=url_format.format(uri="operations/run_provision_task"),
            get_provision_tasks=url_format.format(uri="operations/get_provision_tasks"),
            provision_params=url_format.format(uri="operations/get_provision_params"),
            provision_validation=url_format.format(uri="operations/validate_provision_params"),
            map_existing_volume=url_format.format(uri="operations/map_existing_volume"),
            open_mount_point=url_format.format(uri="operations/open_mount_point"),
            performance_test=url_format.format(uri="volumes/performance/performance_test"),
            cancel_performance_test=url_format.format(uri="volumes/performance/cancel_performance_test"),
            resize_validation=url_format.format(uri="operations/resize/resize_validation"),
            get_resize_tasks=url_format.format(uri="operations/resize/get_resize_tasks"),
            run_resize_task=url_format.format(uri="operations/resize/run_resize_task"),
            query_mount_after_map=url_format.format(uri="operations/query_mount_after_map")
        )
        return sources

    def _get_columns(self):
        return HostVolumesTable.COLUMN_NAMES

    auth = Auth()
    readiness = Readiness()
    registration = HostRegistration()
    volumes = Volumes()
    operations = Operations()
