__import__("pkg_resources").declare_namespace(__name__)

import infi.exceptools
import infi.reboot
from .assertions import AssertionMixin

# pylint: disable=W0710,R0904,W0102,W0622

class CaseResultException(infi.exceptools.InfiException):
    pass

class Result(CaseResultException):
    def __init__(self, message):
        self.message = message

    def get_name(self):
        return self.__class__.__name__

    def __repr__(self):
        return "<{}: {}>".format(self.get_name(), self.message)

    def __str__(self):
        return "{}: {}".format(self.get_name().upper(), self.message)

class Reboot(Result):
    pass

class Warning(Result):
    pass

class Skip(Result):
    pass

class Failure(Result):
    pass

class Error(Result):
    pass

class BaseCase(object):
    failureException = Failure

    def __init__(self, category):
        super(BaseCase, self).__init__()
        self.category = category

    def get_name(self):
        return self.__class__.__name__

    def get_kb_number(self):
        raise NotImplementedError()

    def __repr__(self):
        return "<{}: {}>".format(self.get_name(), self.get_description())

    def __str__(self):
        return "{}: {}".format(self.category, self.get_description())

    def fail(self, message):
        raise Failure(message)

    def skip(self, message):
        raise Skip(message)

    def warn(self, message):
        raise Warning(message)

    def ask_for_reboot(self, reason="Reboot is required for the changes to take effect"):
        infi.reboot.ask_for_reboot(self.get_name())
        raise Reboot(reason)

    def error(self, message="An unexpected error has occurred"):
        raise infi.exceptools.chain(Error(message))

    def has_reboot_took_place(self):
        return infi.reboot.has_reboot_took_place(self.get_name())

class Case(BaseCase, AssertionMixin):

    required_for_provisioning = False

    def get_kb_number(self):
        raise NotImplementedError()

    def get_description(self):
        raise NotImplementedError()

    def run_test(self):
        raise NotImplementedError()

class FixableCase(Case):
    def get_fix_description(self):
        return None

    def run_fix(self):
        raise NotImplementedError()

class Category(object):
    def __init__(self, description, items=[]):
        self._description = description
        self._items = items

    def get_name(self):
        return self.__class__.__name__

    def get_description(self):
        return self._description

    def get_cases(self):
        return self._items

    def add_case(self, item):
        self._items.append(item)

    def add_cases(self, items):
        self._items.extend(items)

    def __repr__(self):
        return "{}: {!r}".format(self.get_name(), self.get_cases())

    def __str__(self):
        return self.get_description()
