from logging import getLogger
from shutil import copy
from infi.recipe.application_packager import utils
from os import path, curdir, makedirs, listdir, remove, getcwd
from infi.vendata.powertools.settings.cases.aix.odm import ODM_FILEPATH, CLEAN_ODM_FILEPATH, ODM_RPM_FILEPATH


logger = getLogger(__name__)

class ODMPackager(object):
    # This is a stripped version of infi.recipe.application_packager.rpm.Recipe

    def build_package(self):
        from re import search
        import platform
        self._files = set()
        self._directories = set()
        self._directories_to_clean = set()
        self._buildout_dir = getcwd()
        self._odm_version = self.get_odm_version()

        with utils.temporary_directory_context() as tempdir:
            self._buildroot = tempdir
            self._put_all_files()
            with utils.chdir(self.get_working_directory()):
                specfile = self._create_specfile()
                output = self._call_rpmbuild(specfile)
                rpm_wrote = search("Wrote: (.*)", output).groups()[0]
                if path.exists(self.rpm_filepath):
                    remove(self.rpm_filepath)
                copy(rpm_wrote, self.rpm_filepath)
            return self.rpm_filepath

    def get_package_name(self):
        return 'infinidat-odm'

    def get_odm_version(self):
        cmd = r"git describe --tags `git log --pretty=format:'%h' {} | head -n1`"
        version, post, commit = utils.execute.execute_assert_success(cmd.format(ODM_FILEPATH), shell=True).get_stdout()[1:].split('-')
        return '{}.{}'.format(version,post)

    def get_working_directory(self):
        return path.join(self._buildout_dir, 'parts')

    def get_install_prefix(self):
        return '/opt/infinidat/odm'

    def _create_specfile(self):
        from jinja2 import Environment, PackageLoader
        env = Environment(loader=PackageLoader('infi.vendata.powertools', 'settings/cases/aix/odm/rpm'))
        template = env.get_template("rpmspec.in")
        kwargs = self._get_template_kwargs()
        content = template.render(kwargs)
        specfile = 'rpm.spec'
        with open(specfile, 'w') as fd:
            fd.write(content)
        return specfile

    def _mkdir(self, name, parent_directory, just_add_it=False):
        dst = path.join(parent_directory, name)
        src = "{}{}/{}".format(self._buildroot, parent_directory, name)
        self._directories.add(dst) if not just_add_it else None
        self._directories_to_clean.add(dst)
        makedirs(src) if not just_add_it else None
        return dst

    def _add_directory(self, src, parent_directory, recursive=True, only_directory_tree=False):
        source_dirname = path.basename(src)
        destination_directory = self._mkdir(source_dirname, parent_directory, only_directory_tree)
        for filename in listdir(src):
            filepath = path.join(src, filename)
            if path.isfile(filepath):
                if filename.endswith('pyc') or only_directory_tree:
                    continue
                self._add_file(filepath, destination_directory)
            elif recursive:
                self._add_directory(filepath, destination_directory, recursive, only_directory_tree)

    def _add_file(self, source_filepath, destination_directory, destination_filename=None):
        if path.dirname(source_filepath) == '':
            source_filepath = path.join(self._buildout_dir, source_filepath)
        if destination_filename is None:
            destination_filename = path.basename(source_filepath)
        buildroot_filepath = "{}{}/{}".format(self._buildroot, destination_directory, destination_filename)
        destination_filepath = path.join(destination_directory, destination_filename)
        copy(source_filepath, buildroot_filepath)
        self._files.add(destination_filepath)

    def _get_template_kwargs(self):
        directories_to_clean = [item for item in self._directories_to_clean]
        directories_to_clean.sort()
        directories_to_clean.reverse()

        return {
           'product_name': 'Infinidat Host PowerTools ODM',
           'product_description': 'Infinidat Host PowerTools ODM',
           'package_name': self.get_package_name(),
           'package_version': self._odm_version,
           'package_arch': 'ppc',
           'target_arch': 'ppc',
           'requires_declaration': '',
           'close_on_upgrade_or_removal' : '0',
           'prefix': self.get_install_prefix(),
           'build_root': self._buildroot,
           'files': "\n".join(self._files),
           'directories': '',
           'directories_to_clean': '',
           'aix': True
           }

    def _put_all_files(self):
        from pkg_resources import resource_filename

        makedirs("{}{}".format(self._buildroot, self.get_install_prefix()))
        self._mkdir(self.get_install_prefix(), path.sep, True)
        self._add_file(ODM_FILEPATH, self.get_install_prefix())
        self._add_file(CLEAN_ODM_FILEPATH, self.get_install_prefix())
        self._add_file(CLEAN_ODM_FILEPATH, self.get_install_prefix())

    def get_rpm_filename(self):
        from infi.os_info import get_platform_string
        return "{}-{}-{}.rpm".format(self.get_package_name(), self._odm_version,
                                     get_platform_string())

    @property
    def rpm_filepath(self):
        return path.join(self.get_working_directory(), self.get_rpm_filename())

    def _call_rpmbuild(self, specfile):
        return utils.execute.execute_assert_success(["rpm", '--verbose', '--buildroot', self._buildroot, '-bb', specfile]).get_stdout()

def build():
    import platform
    if platform.system() == "AIX":
        odm_path = ODMPackager().build_package()
        copy(odm_path, ODM_RPM_FILEPATH)

if __name__ == "__main__":
    build()