from infi.vendata.powertools.settings.case import FixableCase, Failure
from infi.vendata.powertools.utils import is_veritas_multipathing_installed
from os.path import join, abspath, dirname, exists
from infi.pyutils.contexts import contextmanager
from infi.multipathtools.client import DEFAULT_CONF_FILE
from infi.pyutils.retry import retry_func, WaitAndRetryStrategy

MULTIPATH_CONF_PATH = abspath(join(dirname(__file__), "multipath.conf"))

from logging import getLogger
log = getLogger(__name__)

# pylint: disable=R0904,W0201

def regex_match_or_error(*args, **kwargs):
    from re import match, error
    try:
        return match(*args, **kwargs)
    except error:
        log.debug(error)
        return None


def workaround_missing_vid_pid(running_config):
    def _fix_device(device):
        if device.product is None:
            device.product = ''
        if device.vendor is None:
            device.vendor = ''

    for device in running_config.devices:
        _fix_device(device)
    for device in running_config.blacklist.device:
        _fix_device(device)
    for device in running_config.whitelist.device:
        _fix_device(device)
    return running_config


class MultipathConfigBase(FixableCase):
    def _get_recommended_content(self):
        with open(MULTIPATH_CONF_PATH, 'r') as fd:
            return fd.read()

    def _set_up_host_multipath_info(self):
        from infi.multipathtools.config import Configuration
        self.recommended_configuration = Configuration.from_multipathd_conf(self._get_recommended_content())
        self.client = self._get_client()
        self._ensure_configuration_exists()
        self.running_config = self._get_running_config()

    def _set_up_multipath_stuff(self):
        self._set_up_host_multipath_info()
        self.recommended_infinidat_device_settings = self.recommended_configuration.devices[0]
        self.infinidat_wwids = self._get_infinidat_wwids()

    def _ensure_configuration_exists(self):
        from infi.multipathtools.errors import ConnectionError
        if not exists(DEFAULT_CONF_FILE):
            with open(DEFAULT_CONF_FILE, "w"):
                pass
            try:
                self.client.rescan()
            except ConnectionError:
                pass

    def _get_infinidat_wwids(self):
        return self._get_wwids_from_block_devices(self._get_all_infinidat_disks())

    def _assert_on_dict(self, expected, actual, allow_none=False):
        # We use allow_none for cases in which the installed version of multipath-tools accepts this key
        # but does not accept the value we want (e.g. prio alua on redhat-5)
        for key, expected_value in expected.iteritems():
            if expected_value is None:
                continue
            if key not in actual.keys():
                # this key isn't in the existing configuration, nor has a default None value in MultipathClient
                continue
            actual_value = actual.get(key)
            if actual_value is not None:
                actual_value = actual_value.strip('"')
            msg = "Expected value for {} is {!r}, but the actual value is {!r}"
            self.assertIn(actual_value, [expected_value.strip('"')] + ([None] if allow_none else []),
                          msg.format(key, expected_value, actual_value))

    def _get_client(self):  # pragma: no cover
        from infi.multipathtools import MultipathClient
        return MultipathClient()

    def _get_running_config(self):
        from infi.multipathtools.errors import ConnectionError
        try:
            return workaround_missing_vid_pid(self.client.get_multipathd_conf())
        except ConnectionError:
            self.skip("Multipath Daemon must be running for this test to pass")

    def _assert_default_config(self):
        self._assert_on_dict(self.recommended_configuration.attributes, self.running_config.attributes)

    def _get_live_infinidat_device(self):
        expected = (self.recommended_infinidat_device_settings.vendor.strip('"\'.*'),
                    self.recommended_infinidat_device_settings.product.strip('"\'.*'))
        for device in self.running_config.devices:
            actual = (device.vendor.strip('"\'.*'), device.product.strip('"\'.*'))
            if actual == expected:
                return device
        return None

    def _assert_infinidat_device(self):
        live_device = self._get_live_infinidat_device()
        if not is_veritas_multipathing_installed():
            if live_device is None:
                self.fail("There is no configuration for INFINIDAT device")
            self._assert_on_dict(self.recommended_infinidat_device_settings, live_device, allow_none=True)
        else:
            if live_device:
                self.fail("There is a configuration for INFINIDAT device, although veritas multipathing is installed")

    def _fix_defaults(self):
        for key, value in self.recommended_configuration.attributes.iteritems():
            if value is None:
                continue
            self.running_config.attributes[key] = value

    def _fix_device(self):
        device = self._get_live_infinidat_device()
        if not is_veritas_multipathing_installed():
            if device is None:
                from infi.multipathtools.config import HardwareEntry
                device = HardwareEntry()
                self.running_config.devices.insert(0, device)
            for key, value in self.recommended_infinidat_device_settings.iteritems():
                if value is None:
                    continue
                device[key] = value
        else:
            def _filter(dev):
                vid = self.recommended_infinidat_device_settings.vendor.strip('".*')
                pid = self.recommended_infinidat_device_settings.product.strip('".*')
                return not (dev.vendor.strip('".*') == vid and dev.product.strip('".*') == pid)
            self.running_config.devices = filter(_filter, self.running_config.devices)

    def _get_exclusions_by_vid_pid(self):
        def _filter(device):
            product_match = regex_match_or_error(device.product.strip('"'),
                                  self.recommended_infinidat_device_settings.product.strip('"'))
            vendor_match = regex_match_or_error(device.vendor.strip('"'),
                                  self.recommended_infinidat_device_settings.vendor.strip('"'))
            return product_match is not None and vendor_match is not None
        return filter(_filter, self.running_config.blacklist.device)

    def _assert_exclusions_by_vid_pid(self):
        if not is_veritas_multipathing_installed():
            self.assertEqual([], self._get_exclusions_by_vid_pid(),
                             "There is an exclusion rule for INFINIDAT Vendor ID and Product ID")
        else:
            msg = "There isn't an exclusion rule for INFINIDAT Vendor ID and Product ID," + \
            " but veritas multipathing is enabled"
            self.assertNotEqual([], self._get_exclusions_by_vid_pid(), msg)

    def _fix_exclusions_by_vid_pid(self):
        for item in self._get_exclusions_by_vid_pid():
            self.running_config.blacklist.device.remove(item)

    def _add_infinidat_exclusions_by_vid_pid(self):
        from infi.multipathtools.config import Device
        new_exclude = Device()
        vid = self.recommended_infinidat_device_settings.vendor.strip('"')
        pid = self.recommended_infinidat_device_settings.product.strip('"')
        new_exclude.vendor, new_exclude.product = vid, pid
        self.running_config.blacklist.device.append(new_exclude)

    def _get_exclusions_by_regex(self):
        def _filter(item):
            """:returns: True if the filter excludes sd devices"""
            string = "sda"
            pattern = item.strip('"')
            pattern = '.*' if pattern == '*' else pattern
            result = regex_match_or_error(pattern, string) is not None
            log.debug("regex pattern matching for {} on {} returned {!r}".format(pattern,
                                                                                 string,
                                                                                 result))
            return result
        return filter(_filter, self.running_config.blacklist.devnode)

    def _fix_exclusions_by_regex(self):
        for item in self._get_exclusions_by_regex():
            self.running_config.blacklist.devnode.remove(item)

    def _assert_exclusions_by_regex(self):
        self.assertEqual([], self._get_exclusions_by_regex(),
                         "There is exlcusion rule for sd devices")

    def _get_all_infinidat_disks(self):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_scsi_block_devices
        from infi.storagemodel.errors import RescanIsNeeded
        try:
            infininidat_disks = get_infinidat_scsi_block_devices()
        except RescanIsNeeded:
            log.exception("rescan is needed")
            return []
        return infininidat_disks

    def _get_wwids_from_block_devices(self, block_devices):
        from infi.storagemodel.errors import DeviceDisappeared
        wwids = []
        for device in block_devices:
            try:
                wwids.append(device.get_scsi_serial_number())
            except DeviceDisappeared:
                log.exception("rescan is needed")
        return wwids

    def _get_exclusions_by_sn(self):
        def _filter(wwid_pattern):
            """":returns:True if the wwid patern matches any infinidat wwid"""
            wwid_pattern = wwid_pattern.strip('"')
            wwid_pattern = '.*' if wwid_pattern == '*' else wwid_pattern
            return any([regex_match_or_error(wwid_pattern, string) is not None for string in self.infinidat_wwids])
        return filter(_filter, self.running_config.blacklist.wwid)

    def _fix_exclusions_by_sn(self):
        for item in self._get_exclusions_by_sn():
            self.running_config.blacklist.wwid.remove(item)

    def _assert_exclusions_by_sn(self):
        self.assertEqual([], self._get_exclusions_by_sn(),
                         "There is exlcusion rule for sg devices")

    def _prepare_fix(self):
        """ This function is called when the case fails, regardless of the flags called
        (If --fix or --auto-fix were passed, it don't matter).
        It prepares the backup filepath for later, and creates the files:
        * multipath.conf.{timestamp}.full (current configuration with all default values)
        * multipath.conf.{timestamp}.fixed (same after the case fixes)
        This means we run the fixes to the configuration anyway. The fixes are done by the subclasses' _run_conf_fixes
        method."""
        from time import strftime
        from shutil import copy
        path_with_time = DEFAULT_CONF_FILE + "_{}".format(strftime("%Y_%m_%d_%H_%M_%S"))
        self._conf_backup_path = path_with_time + ".bak"
        self._conf_full_path = path_with_time + ".full"
        self._conf_fixed_path = path_with_time + ".fixed"
        if not exists(DEFAULT_CONF_FILE):
            self._conf_backup_path = None
        else:
            copy(DEFAULT_CONF_FILE, self._conf_backup_path)
        self.client.write_to_multipathd_conf(self.running_config, filepath=self._conf_full_path)
        self._run_conf_fixes()
        self.client.write_to_multipathd_conf(self.running_config, filepath=self._conf_fixed_path)

    @contextmanager
    def _prepare_fix_if_needed(self):
        try:
            yield
        except Failure:
            self._prepare_fix()
            raise

    def _run_fix(self):
        """ base method for all subclasses' run_fix """
        from infi.vendata.powertools.settings.cases.linux.base_cases.boot_from_san import mark_generation_of_initrd_as_needed
        from shutil import copy
        from os import remove
        mark_generation_of_initrd_as_needed()
        copy(self._conf_fixed_path, DEFAULT_CONF_FILE)
        remove(self._conf_fixed_path)
        remove(self._conf_full_path)
        self._ensure_fixed()

    @retry_func(WaitAndRetryStrategy(max_retries=3, wait=5))
    def _ensure_fixed(self):
        self.client.rescan()
        self.running_config = self._get_running_config()
        self._assert_fixed()

    def _assert_fixed(self):
        raise NotImplementedError

    def _get_fix_description_end(self):
        result = "\nyou can review the new configuration at {}\n"
        if self._conf_backup_path:
            result += "current configuration is backed up at {}\n".format(self._conf_backup_path)
        return result.format(self._conf_fixed_path)

class MultipathConfigGlobal(MultipathConfigBase):
    def run_test(self):  # pragma: no cover
        self._set_up_multipath_stuff()
        with self._prepare_fix_if_needed():
            self._assert_fixed()

    def run_fix(self):  # pragma: no cover
        self._run_fix()

class MultipathConfigDefaults(MultipathConfigGlobal):

    def _assert_fixed(self):
        self._assert_default_config()

    def _run_conf_fixes(self):
        self._fix_defaults()

    def get_description(self):
        return "Checking global parameters"

    def get_fix_description(self):
        return "override the values of the defaults section of the configuration file with recommended values." + \
            self._get_fix_description_end()

    def get_kb_number(self):
        return "202319232"

class MultipathConfigRecommened(MultipathConfigGlobal):

    def _assert_fixed(self):
        self._assert_infinidat_device()

    def _run_conf_fixes(self):
        self._fix_device()

    def get_fix_description(self):
        return "create and set the recommended values for InfiniBox." + self._get_fix_description_end()

    def get_description(self):
        return "Checking InfiniBox Device Settings"

    def get_kb_number(self):
        return "202404151"

class MultipathConfigInfinidatExclusions(MultipathConfigGlobal):

    required_for_provisioning = True

    def _assert_fixed(self):
        self._assert_exclusions_by_vid_pid()
        self._assert_exclusions_by_regex()
        self._assert_exclusions_by_sn()

    def _run_conf_fixes(self):
        self._fix_exclusions_by_vid_pid()
        self._fix_exclusions_by_regex()
        self._fix_exclusions_by_sn()
        if is_veritas_multipathing_installed():
            self._add_infinidat_exclusions_by_vid_pid()

    def get_fix_description(self):
        return "remove blacklist rules that block InfiniBox devices from the multipath configuration file." + \
            self._get_fix_description_end()

    def run_fix(self):  # pragma: no cover
        self._run_fix()

    def get_kb_number(self):
        return "202319242"

    def get_description(self):
        return "Checking for InfiniBox device exclusions"

class MultipathConfigLocalExclusions(MultipathConfigBase):

    def _assert_fixed(self):
        self.assertEqual([], self._get_non_blacklisted_local_devices_vid_pid(),
                             "Local disks are not excluded in multipath.conf")

    def _scsi_serial_exists(self, device):
        try:
            return bool(device.get_scsi_serial_number())
        except:
            log.debug("no scsi serial for local device {!r}, so no need to exclude it".format(device))

    def _get_local_devices_vid_pid(self):
        from infi.storagemodel import get_storage_model
        from infi.storagemodel.connectivity import LocalConnectivity
        model = get_storage_model()
        model.refresh()
        local_block_devices = model.get_scsi().get_all_scsi_block_devices()
        vid_pid_set = set()
        for device in local_block_devices:
            if isinstance(device.get_connectivity(), LocalConnectivity):
                if not self._scsi_serial_exists(device):
                    log.debug("no scsi serial for local device {!r}, so no need to exclude it".format(device))
                    continue
                try:
                    value = device.get_scsi_vid_pid()
                except:
                    log.exception("failed to get scsi vid/pid for {!r}".format(device))
                else:
                    log.debug("found local device {!r}: {!r}".format(device, value))
                    vid_pid_set.add(value)
            else:
                log.debug("no local connectivity for device {!r}: {!r}".format(device, device.get_connectivity()))

        for vid_pid in vid_pid_set:
            quoted_vid_pid = ['"{}"'.format(component) for component in vid_pid]
            yield quoted_vid_pid

    def _get_non_blacklisted_local_devices_vid_pid(self):
        # get vid and pid of local devices that are not blacklisted in the Multipath configuration
        blacklisted_vid_pids = [(dev.vendor, dev.product) for dev in self.running_config.blacklist.device]
        return [(dev_vid, dev_pid) for (dev_vid, dev_pid) in self._get_local_devices_vid_pid() if
                (dev_vid, dev_pid) not in blacklisted_vid_pids]

    def run_test(self):  # pragma: no cover
        self._set_up_multipath_stuff()
        self.client.rescan()
        with self._prepare_fix_if_needed():
            self._assert_fixed()

    def get_fix_description(self):
        return "add blacklist rules that excludes local devices to the multipath configuration file." + \
            self._get_fix_description_end()

    def _run_conf_fixes(self):
        from infi.multipathtools.config import Device
        for dev_vid, dev_pid in set(self._get_non_blacklisted_local_devices_vid_pid()):
            new_exclude = Device()
            new_exclude.vendor, new_exclude.product = dev_vid, dev_pid
            self.running_config.blacklist.device.append(new_exclude)

    def run_fix(self):  # pragma: no cover
        self._run_fix("local exclusions")
        self.ask_for_reboot("Reboot is required for the changes to take effect")

    def get_kb_number(self):
        return "202404331"

    def get_description(self):
        return "Checking for local device exclusions"
