from infi.vendata.powertools.settings.case import Case
from . import MultipathConfigBase, regex_match_or_error

from logging import getLogger

logger = getLogger(__name__)

POWERPATH_PATH = '/sbin/powermt'
POWERPATH_PACKAGE = 'EMCpower.LINUX.x86_64'

class PowerPath(object):
    def _get_ignore_flag(self):
        return "ignore_powerpath"

    def is_powerpath_ignore_flag(self):
        ignore_flag = self._get_ignore_flag()
        if ignore_flag is not None:
            from infi.vendata.powertools.conf import get_configuration
            if ignore_flag in get_configuration().get('readiness.cases.coexistence'):
                return get_configuration().get('readiness.cases.coexistence.{}'.format(ignore_flag))
        return False

    def is_powerpath_installed(self):
        from infi.pkgmgr import get_package_manager
        from os import path
        return path.exists(POWERPATH_PATH) or get_package_manager().is_package_installed(POWERPATH_PACKAGE)

    def should_fail(self):
        return self.is_powerpath_installed() and not self.is_powerpath_ignore_flag()

    def should_warn(self):
        return self.is_powerpath_installed() and self.is_powerpath_ignore_flag()


class PowerPathCase(Case, PowerPath):
    def get_kb_number(self):
        return "207217755"

    def get_description(self):
        return "PowerPath coexistence with multipathd"

    def run_test(self):
        msg = "PowerPath does not co-exist with linux native multipath"
        if self.should_fail():
            self.fail(msg)
        elif self.should_warn():
            self.warn(msg)


class PowerPathExclusions(MultipathConfigBase, PowerPath):

    required_for_provisioning = True

    def _get_emc_exclusions_by_vid_pid(self, config):
        def _filter(device):
            vendor_match = regex_match_or_error(device.vendor.strip('"'), "EMC")
            return vendor_match is not None and (device.product.strip('"') == ".*")
        return filter(_filter, config.blacklist.device)

    def _is_multipath_conf_exists(self):
        from os import path
        from infi.multipathtools.client import DEFAULT_CONF_FILE
        return path.exists(DEFAULT_CONF_FILE)

    def _get_static_multipath_conf(self):
        from infi.multipathtools.client import DEFAULT_CONF_FILE
        from infi.multipathtools.config import Configuration
        if self._is_multipath_conf_exists():
            return Configuration().from_multipathd_conf(open(DEFAULT_CONF_FILE, 'rb').read())
        else:
            return Configuration()

    def _set_static_multipath_conf(self, config):
        from infi.multipathtools.client import DEFAULT_CONF_FILE
        with open(DEFAULT_CONF_FILE, 'wb') as multipath_conf_file:
            multipath_conf_file.write(config.to_multipathd_conf())

    def _is_emc_blacklisted(self, config):
        return not self._get_emc_exclusions_by_vid_pid(config) == []

    def _is_multipath_running(self):
        from infi.sysv_service import LinuxInitService
        service = LinuxInitService("multipathd", "multipathd")
        return service.is_running()

    def run_test(self):
        if self.is_powerpath_installed():
            if not self._is_multipath_conf_exists() or not self._is_emc_blacklisted(self._get_static_multipath_conf()):
                self.fail("There is no exclusion rule for all products of EMC Vendor")
            # Extra check in case multipathd is running
            if self._is_multipath_running():
                self._set_up_host_multipath_info()
                self.assertTrue(self._is_emc_blacklisted(self.running_config),
                                 "There is no exclusion rule for all products of EMC Vendor")
        else:
            pass

    def get_fix_description(self):
        return "add EMC products to the exclusion list in the multipathd configuration file."

    def _add_emc_exclusions_by_vid_pid(self, config):
        from infi.multipathtools.config import Device
        new_exclude = Device()
        new_exclude.vendor = "EMC"
        new_exclude.product = ".*"
        config.blacklist.device.append(new_exclude)

    def _run_conf_fixes(self):
        self._add_emc_exclusions_by_vid_pid(self.running_config)

    def run_fix(self):
        if not self._is_multipath_running():
            config = self._get_static_multipath_conf()
            self._add_emc_exclusions_by_vid_pid(config)
            self._set_static_multipath_conf(config)
        else:
            self._prepare_fix()
            self._run_fix("powerpath")

    def get_kb_number(self):
        return "207217755"

    def get_description(self):
        return "Checking that EMC products are excluded in multipathd configuration"
