import os
import sys
import json
from mock import patch, MagicMock
from infi.pyutils.contexts import contextmanager
from unittest import TestCase, SkipTest
from fake_filesystem_unittest import TestCase as FakeFSTestCase
from fake_filesystem import FakeOsModule, FakeFileOpen
from infi.vendata.powertools.conf import DEFAULTS_FILEPATH
from infi.vendata.powertools.settings import case
from .. import fc_driver_attributes


MODPROBE_D_PATH = '/etc/modprobe.d'
INFINIDAT_CONF_PATH = os.path.join(MODPROBE_D_PATH, 'infinidat.conf')
MODPROBE_CONF_PATH = '/etc/modprobe.conf'
CONFLICT_CONTENTS = "options lpfc lpfc_hba_queue_depth=666"
NON_CONFLICT_CONTENTS_BECAUSE_DRV_NAME = "options zulu lpfc_hba_queue_depth=666"
NON_CONFLICT_CONTENTS_BECAUSE_OPTION = "options lpfc zulu=666"
CS514_CONTENTS = open(os.path.join('tests', 'CS-541', 'etc', 'modprobe.conf.20140915')).read()


@contextmanager
def fake_os_context(linux_dist, linux_ver):
    def empty_side_effect(*args, **kwargs):
        pass
    def side_effect(*args, **kwargs):
        return linux_dist, linux_ver, None
    with patch('infi.vendata.powertools.settings.utils.backup_and_create_initrd', MagicMock(side_effect=empty_side_effect)):
        with patch('infi.vendata.powertools.settings.case.FixableCase.ask_for_reboot', MagicMock(side_effect=empty_side_effect)):
            with patch('infi.vendata.powertools.settings.cases.linux.base_cases.boot_from_san.ask_for_reboot', MagicMock(side_effect=empty_side_effect)):
                with patch('infi.vendata.powertools.settings.cases.linux.base_cases.fc_driver_attributes.reloaded_fc_drivers', MagicMock(side_effect=empty_side_effect)):
                    with patch('platform.linux_distribution', MagicMock(side_effect=side_effect)):
                        yield


@contextmanager
def redhat7_context():
    with fake_os_context("Red Hat Enterprise Linux Server", "7.0"):
        yield


@contextmanager
def redhat6_context():
    with fake_os_context("Red Hat Enterprise Linux Server", "6.5"):
        yield


@contextmanager
def redhat5_context():
    with fake_os_context("Red Hat Enterprise Linux Server", "5.6"):
        yield

@contextmanager
def centos5_context():
    with fake_os_context("CentOs", "5.6"):
        yield

@contextmanager
def ubuntu14_context():
    with fake_os_context("Ubuntu", "14.04"):
        yield


class TestDifferentDistros(FakeFSTestCase):
    @classmethod
    def setUpClass(cls):
        cls._save_source_files_for_pdb()

    def setUp(self):
        from platform import system
        if 'windows' in system().lower():
            self.skipTest("Skipping on windows since it's a linux module")
        default_conf_data = open(DEFAULTS_FILEPATH, 'r').read()
        self.setUpPyfakefs()
        self.fs.CreateFile(DEFAULTS_FILEPATH, contents=default_conf_data)
        self.fs.CreateDirectory('/etc')
        self.fs.CreateDirectory('/sys/module/lpfc')
        self.fs.CreateDirectory('/sys/module/qla2xxx')
        self._put_source_files_in_fakefs()
        self._drivers_attributes = json.load(open(DEFAULTS_FILEPATH))['readiness']['cases']['fc_driver_attributes']

        # Create fake driver elfs
        for driver_name, attrs in self._drivers_attributes.iteritems():
            driver_path = os.path.join("/lib", "modules", os.uname()[2], "kernel", "drivers", "scsi", driver_name, driver_name+'.ko')
            driver_elf_data = '\n'.join(['parm={}'.format(attr) for attr in attrs.iterkeys()])
            self.fs.CreateFile(driver_path, contents=driver_elf_data)

        # For some reason glob is not replaced
        self._original_glob = sys.modules['glob']
        sys.modules['glob'] = self.stubber.fake_glob

    @classmethod
    def _save_source_files_for_pdb(cls):
        cls._module_data = {}
        modules = dict(sys.modules)
        for module_name, module in modules.iteritems():
            if module and hasattr(module, '__file__'):
                filename = module.__file__.replace('.pyc', '.py')
                filename = filename.replace('.pyo', '.py')
                try:
                    data = open(filename, 'r').read()
                    cls._module_data[module.__file__] = data
                    cls._module_data[filename] = data
                except:
                    pass

    def _put_source_files_in_fakefs(self):
        for module, data in self._module_data.iteritems():
            self.fs.CreateFile(module, contents=data)

    def _assert_file_nonexistent_or_empty(self, filename):
        self.assertTrue(not FakeOsModule(self.fs).path.exists(filename) or
                        FakeFileOpen(self.fs)(filename).read() == '')

    def _assert_file_exists_and_nonempty(self, filename):
        self.assertTrue(FakeOsModule(self.fs).path.exists(filename) and
                        FakeFileOpen(self.fs)(filename).read() != '')

    def tearDown(self):
        sys.modules['glob'] = self._original_glob
        self.tearDownPyfakefs()

    def test_redhat7__nonexistent_conf(self):
        with redhat7_context():
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_ubuntu14__acts_like_redhat7(self):
        with ubuntu14_context():
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_redhat6__nonexistent_conf(self):
        with redhat6_context():
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_D_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat5__nonexistent_conf(self):
        with redhat5_context():
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_D_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat7__nondir_modprobe_d(self):
        with redhat7_context():
            self.fs.CreateFile(MODPROBE_D_PATH, contents="ohno!")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Failure):
                test_case.run_fix()

    def test_redhat7__conflict(self):
        with redhat7_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.conf'), contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat6__conflict_in_output_file(self):
        with redhat6_context():
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_exists_and_nonempty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat6__conflict_in_modprob_conf(self):
        with redhat6_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'noconflict.conf'), contents="bla")
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_exists_and_nonempty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat5__conflict_in_output_file(self):
        with redhat5_context():
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_exists_and_nonempty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat5__conflict_in_non_conf_ext(self):
        with redhat5_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.crap'), contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat5__modprobed_ignored(self):
        with redhat5_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.conf'), contents=CONFLICT_CONTENTS)
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents="nonconflict")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)

    def test_redhat5__modprobed_notcreated(self):
        # This test makes sure modprobe.d is not created (Because if it is created on redhat 5, all the settings in modprobe.conf will be igonred)
        with redhat5_context():
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents="nonconflict")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_D_PATH)

    def test_redhat5__conflict_in_modprobed(self):
        with redhat5_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.conf'), contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def _basic_include_test(self):
            self.fs.CreateFile(os.path.join('/', 'conflict.crap'), contents=CONFLICT_CONTENTS)
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'some_file'), contents="include /conflict.crap\n")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_redhat5__conflict_in_include_file(self):
        with redhat5_context():
            self._basic_include_test()

    def test_centos5__acts_like_redhat5(self):
        with centos5_context():
            self._basic_include_test()

    def test_redhat5__two_include_files_first_nonexistent(self):
        with redhat5_context():
            self.fs.CreateFile(os.path.join('/', 'conflict.crap'), contents=CONFLICT_CONTENTS)
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'some_file'), contents="asdf\ninclude nonexistent\ninclude /conflict.crap\n")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()
            self._assert_file_nonexistent_or_empty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_redhat5__second_level_include(self):
        with redhat5_context():
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents="asdf\ninclude /etc/modprobe.d/bar.crap\n")
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'bar.crap'), contents='include /etc/modprobe.d/foo.crap')
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()

    def test_redhat5__include_file_nonexistent(self):
        with redhat5_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'some_file'), contents="include nonexist.crap")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_redhat6__include_ignored(self):
        with redhat6_context():
            self.fs.CreateFile(os.path.join('/', 'conflict.crap'), contents=CONFLICT_CONTENTS)
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'some_file.conf'), contents="include /conflict.crap\n")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_redhat7__include_ignored(self):
        with redhat7_context():
            self.fs.CreateFile(os.path.join('/', 'conflict.crap'), contents=CONFLICT_CONTENTS)
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'some_file.conf'), contents="include /conflict.crap\n")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)

    def test_redhat6__conflict_in_non_conf_ext(self):
        with redhat6_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.crap'), contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)

    def test_redhat7__conflict_in_non_conf_ext(self):
        with redhat7_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.crap'), contents=CONFLICT_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_nonexistent_or_empty(MODPROBE_CONF_PATH)
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)

    def test_redhat7__non_conflict_because_option(self):
        with redhat7_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.conf'), contents=NON_CONFLICT_CONTENTS_BECAUSE_OPTION)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)

    def test_redhat7__non_conflict_because_drv_name(self):
        with redhat7_context():
            self.fs.CreateFile(os.path.join(MODPROBE_D_PATH, 'conflict.conf'), contents=NON_CONFLICT_CONTENTS_BECAUSE_DRV_NAME)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(INFINIDAT_CONF_PATH)

    def test_redhat6__nondir_modprobe_d(self):
        with redhat6_context():
            self.fs.CreateFile(MODPROBE_D_PATH, contents="ohno!")
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            test_case.run_fix()
            self._assert_file_exists_and_nonempty(MODPROBE_CONF_PATH)

    def test_cs_514(self):
        with redhat5_context():
            self.fs.CreateFile(MODPROBE_CONF_PATH, contents=CS514_CONTENTS)
            test_case = fc_driver_attributes.FiberChannelDriverProperties(None)
            with self.assertRaises(case.Warning):
                test_case.run_fix()


class ReloadFcDriversTestCase(TestCase):
    def test_all_reloaded(self):
        from infi.vendata.powertools.settings.cases.linux.base_cases import fc_driver_attributes
        with patch("infi.vendata.powertools.settings.cases.linux.base_cases.kernel.execute_command"):
            self.assertTrue(fc_driver_attributes.reloaded_fc_drivers())

    def test_none_loaded_to_begin_with(self):
        from infi.vendata.powertools.settings.cases.linux.base_cases import fc_driver_attributes
        _dict = fc_driver_attributes.FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS.copy()
        fc_driver_attributes.FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS.clear()
        self.addCleanup(lambda: fc_driver_attributes.FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS.update(_dict))
        self.assertTrue(fc_driver_attributes.reloaded_fc_drivers())

    def test_one_failed_to_reload(self):
        from infi.vendata.powertools.settings.cases.linux.base_cases import fc_driver_attributes
        from mock import MagicMock
        reloaded_modules = set()

        def side_effect(*args, **kwargs):
            if args == (['/sbin/lsmod'],):
                result = MagicMock()
                result.get_stdout.return_value = "qla2xxx\nlpfc\nbfa\n"
                return result
            if args == (['rmmod', 'lpfc'],):
                raise RuntimeError()
            if args[0][0] == "modprobe":
                reloaded_modules.add(args[0][1])
        with patch("infi.vendata.powertools.settings.cases.linux.base_cases.kernel.execute_command") as execute_command:
            execute_command.side_effect = side_effect
            self.assertFalse(fc_driver_attributes.reloaded_fc_drivers())
        self.assertEquals(reloaded_modules, set(["qla2xxx", "bfa"]))


class FakeRcModulesScript(fc_driver_attributes.RcModulesScript):
    DST = os.path.join(os.path.sep, 'tmp', 'rc.modules.infinidat')
    DST_LINK = os.path.join(os.path.sep, 'tmp', 'rc.modules')


class RcModulesScriptTestCase(TestCase):
    @classmethod
    def setUpClass(self):
        if sys.platform in ('win32', 'darwin', 'sunos5', 'aix7'):
            raise SkipTest("no lsmod")

    def setUp(self):
        self._case = FakeRcModulesScript(None)
        if os.path.exists(self._case.DST_LINK):
            os.remove(self._case.DST_LINK)

    def test_dst_link_does_not_exists(self):
        with self.assertRaises(case.Failure):
            self._case.run_test()

    def test_dst_link_exists_as_regular_file(self):
        with open(self._case.DST_LINK, 'w'):
            pass
        with self.assertRaises(case.Failure):
            self._case.run_test()

    def test_dst_link_exists_but_points_to_a_different_link(self):
        os.symlink(os.path.abspath(os.path.join('buildout.cfg')), self._case.DST_LINK)
        with self.assertRaises(case.Failure):
            self._case.run_test()

    def test_dst_link_points_properly_to_our_file(self):
        os.symlink(self._case.DST, self._case.DST_LINK)
        self._case.run_test()

    def test_fix_creates_link(self):
        with self.assertRaises(case.Failure):
            self._case.run_test()
        self._case.run_fix()
        self._case.run_test()

    def test_fix_fails_if_link_exists_and_invalid(self):
        os.symlink(os.path.abspath(os.path.join('buildout.cfg')), self._case.DST_LINK)
        with self.assertRaises(case.Failure):
            self._case.run_fix()
