from infi.vendata.powertools.settings.case import FixableCase
from infi.os_info import get_platform_string
from logging import getLogger
from os import path
logger = getLogger(__name__)


def get_targetdir():
    platform = get_platform_string()
    linux, distro, version = platform.split('-', 3)[:3]
    if not platform.startswith('linux'):
        raise NotImplementedError()
    if distro in ('redhat', 'centos', 'oracle') and version in ('7', ):
        return path.join(path.sep, 'lib', 'udev', 'rules.d')
    if distro in ('ubuntu') and version in ('trusty', ):
        return path.join(path.sep, 'lib', 'udev', 'rules.d')
    return path.join(path.sep, 'etc', 'udev', 'rules.d')


def _get_resource_filepath(basename):
    return path.abspath(path.join(path.dirname(__file__), basename))


def _get_config():
    from infi.vendata.powertools.conf import get_configuration
    return get_configuration().get('readiness.cases.io_scheduler')


def _get_expected_scheduler_name():
    return _get_config().get('scheduler') or 'noop'


def _get_udev_rule_priority():
    return str(_get_config().get('udev-rule-priority') or 41)


def _get_systemd_rule_priority():
    return str(_get_config().get('systemd-rule-priority') or 99)


def get_src_filepath():
    platform = get_platform_string()
    linux, distro, version = platform.split('-', 3)[:3]
    if not platform.startswith('linux'):
        raise NotImplementedError()
    if distro in ('redhat', 'centos', 'oracle') and version in ('7', ):
        return _get_resource_filepath("99-infinidat-io-scheduler.rules")
    if distro in ('suse') and version in ('12', ):
        return _get_resource_filepath("99-infinidat-io-scheduler.rules")
    if distro in ('ubuntu') and version in ('trusty', ):
        return _get_resource_filepath("99-infinidat-io-scheduler.rules")
    return _get_resource_filepath("41-infinidat-io-scheduler.rules")


def get_dst_filepath():
    basename = path.basename(get_src_filepath())
    if basename.startswith('41'):
        basename = basename.replace('41', _get_udev_rule_priority())
    elif basename.startswith('99'):
        basename = basename.replace('99', _get_systemd_rule_priority())
    return path.join(get_targetdir(), basename)


def get_io_scheduler_for_device(device):
    from re import search
    sysfs = path.join(path.sep, 'sys', 'block', path.basename(device), 'queue', 'scheduler')
    with open(sysfs) as fd:
        contents = fd.read().strip()
    return search(r'(?:\[)(?P<scheduler>[a-z]+)(?:\])', contents).group('scheduler')


def set_io_scheduler_for_device(device):
    sysfs = path.join(path.sep, 'sys', 'block', path.basename(device), 'queue', 'scheduler')
    with open(sysfs, 'w') as fd:
        fd.write(_get_expected_scheduler_name())


def iter_infinidat_single_path_block_devices():
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_scsi_block_devices
    for device in get_infinidat_scsi_block_devices():
        access_path = device.get_block_access_path()
        if path.exists(access_path):
            yield access_path


def set_scheduler_for_current_devices():
    for access_path in iter_infinidat_single_path_block_devices():
        scheduler = get_io_scheduler_for_device(access_path)
        if scheduler != _get_expected_scheduler_name():
            set_io_scheduler_for_device(access_path)


def install_the_udev_rule():
    dst = get_dst_filepath()
    src = get_src_filepath()
    with open(src) as fd:
        contents = fd.read()
    with open(dst, 'w') as fd:
        fd.write(contents)


class IO_Scheduler_Rule(FixableCase):
    required_for_provisioning = False

    def get_kb_number(self):
        return "202404211"

    def get_description(self):
        msg = "Checking that a udev rule exists for setting the '{}' I/O scheduler for NFINIDAT scsi-disks"
        return msg.format(_get_expected_scheduler_name())

    def get_fix_description(self):
        return """Install udev rule for setting the I/O scheduler for new NFINIDAT block devices"""

    def run_fix(self):
        install_the_udev_rule()

    def run_test(self):
        dst = get_dst_filepath()
        src = get_src_filepath()

        def test_rule_exists():
            if not path.exists(dst):
                self.fail("udev rule {} does not exist".format(dst))

        def test_rule_contents():
            with open(src) as sfd:
                with open(dst) as dfd:
                    if sfd.read() != dfd.read():
                        self.fail("contents of udev rule {} is not as expected".format(dst))

        test_rule_exists()
        test_rule_contents()


class IO_Scheduler_AffectedSettings(IO_Scheduler_Rule):
    def get_description(self):
        msg = "Checking that the current I/O scheduler for all NFINIDAT scsi-disks is '{}'"
        return msg.format(_get_expected_scheduler_name())

    def get_fix_description(self):
        return """Set the I/O scheduler for all the current attached devices"""

    def run_test(self):
        def test_current_schedulers():
            for access_path in iter_infinidat_single_path_block_devices():
                scheduler = get_io_scheduler_for_device(access_path)
                if scheduler != _get_expected_scheduler_name():
                    self.fail("some INFINIDAT devices have the wrong I/O scheduler; for example, {} has {}".format(access_path, scheduler))

        test_current_schedulers()

    def run_fix(self):
        set_scheduler_for_current_devices()
