from infi.vendata.powertools.settings.cases.linux.base_cases import io_scheduler
from infi.vendata.powertools.settings import case
import fake_filesystem_unittest
from unittest import SkipTest
from infi import unittest

import mock
import os

SYSTEMD_STYLE_PLATFORMS = ["linux-centos-7-x64", "linux-redhat-7-x64", "linux-ubuntu-trusty-x86"]
OLD_STYLE_PLATFORMS = ["linux-redhat-6-x64", "linux-redhat-5-x86", "linux-suse-11-x64"]


def _get_fake_config():
    return dict(scheduler='noop')


class TargetTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if os.name == 'nt':
            raise SkipTest("this test case relies on posix file paths")

    @unittest.parameters.iterate("platform_string", SYSTEMD_STYLE_PLATFORMS)
    def test_lib_udev(self, platform_string):
        with mock.patch.object(io_scheduler, 'get_platform_string') as get_platform_string:
            get_platform_string.return_value = platform_string
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self.assertEquals(io_scheduler.get_dst_filepath(), "/lib/udev/rules.d/99-infinidat-io-scheduler.rules")

    @unittest.parameters.iterate("platform_string", OLD_STYLE_PLATFORMS)
    def test_etc(self, platform_string):
        with mock.patch.object(io_scheduler, 'get_platform_string') as get_platform_string:
            get_platform_string.return_value = platform_string
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self.assertEquals(io_scheduler.get_dst_filepath(), "/etc/udev/rules.d/41-infinidat-io-scheduler.rules")


class SourceTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if os.name == 'nt':
            raise SkipTest("this test case relies on posix file paths")

    @unittest.parameters.iterate("platform_string", OLD_STYLE_PLATFORMS)
    def test_legacy_rule(self, platform_string):
        with mock.patch.object(io_scheduler, 'get_platform_string') as get_platform_string:
            with mock.patch.object(io_scheduler, '_get_resource_filepath') as _get_resource_filepath:
                _get_resource_filepath.return_value = "/41-infinidat-io-scheduler.rules"
                get_platform_string.return_value = platform_string
                self.assertEquals(io_scheduler.get_src_filepath(), "/41-infinidat-io-scheduler.rules")

    @unittest.parameters.iterate("platform_string", SYSTEMD_STYLE_PLATFORMS)
    def test_systemd_rule(self, platform_string):
        with mock.patch.object(io_scheduler, 'get_platform_string') as get_platform_string:
            with mock.patch.object(io_scheduler, '_get_resource_filepath') as _get_resource_filepath:
                _get_resource_filepath.return_value = "/99-infinidat-io-scheduler.rules"
                get_platform_string.return_value = platform_string
                self.assertEquals(io_scheduler.get_src_filepath(), "/99-infinidat-io-scheduler.rules")


class SchedulerTestCase(fake_filesystem_unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if os.name == 'nt':
            raise SkipTest("this test case relies on posix file paths")

    def setUp(self):
        self.setUpPyfakefs()

    def tearDown(self):
        self.tearDownPyfakefs()

    def test_getter__deadline(self):
        self.fs.CreateDirectory("/sys/block/sda/queue")
        self.fs.CreateFile("/sys/block/sda/queue/scheduler",
                           contents="noop [deadline] cfq")
        self.assertEquals(io_scheduler.get_io_scheduler_for_device("sda"), "deadline")

    def test_getter__noop(self):
        self.fs.CreateDirectory("/sys/block/sda/queue")
        self.fs.CreateFile("/sys/block/sda/queue/scheduler",
                           contents="[noop] deadline cfq")
        self.assertEquals(io_scheduler.get_io_scheduler_for_device("sda"), "noop")

    def test_setter(self):
        self.fs.CreateDirectory("/sys/block/sda/queue")
        with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
            io_scheduler.set_io_scheduler_for_device("sda")


class FakeMixin(object):
    def _setup_rule(self, src_contents, dst_contents):
        src = io_scheduler.get_src_filepath()
        dst = io_scheduler.get_dst_filepath()
        self.fs.CreateDirectory(os.path.dirname(src))
        self.fs.CreateFile(src, contents=src_contents)
        self.fs.CreateDirectory(os.path.dirname(dst))
        self.fs.CreateFile(dst, contents=dst_contents)


class AssertionsTestCase(fake_filesystem_unittest.TestCase, FakeMixin):
    @classmethod
    def setUpClass(cls):
        if os.name == 'nt':
            raise SkipTest("this test case relies on posix file paths")

    def setUp(self):
        self.setUpPyfakefs()
        patch = mock.patch.object(io_scheduler, 'get_platform_string', new=lambda: 'linux-redhat-6')
        patch.__enter__()
        self.addCleanup(patch.__exit__, (None, None, None))

    def tearDown(self):
        self.tearDownPyfakefs()

    def test_rule_not_installed(self):
        task = io_scheduler.IO_Scheduler_Rule(mock.MagicMock())
        with self.assertRaises(case.Failure) as cm, mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
            task.run_test()
        self.assertEquals(cm.exception.message,
                          "udev rule /etc/udev/rules.d/41-infinidat-io-scheduler.rules does not exist")

    def test_rule_not_installed_redhat_7(self):
        task = io_scheduler.IO_Scheduler_Rule(mock.MagicMock())
        with self.assertRaises(case.Failure) as cm, mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
            with mock.patch.object(io_scheduler, 'get_platform_string', new=lambda: 'linux-redhat-7'):
                task.run_test()
        self.assertEquals(cm.exception.message,
                          "udev rule /lib/udev/rules.d/99-infinidat-io-scheduler.rules does not exist")

    def _setup_rule(self, src_contents, dst_contents):
        src = io_scheduler.get_src_filepath()
        dst = io_scheduler.get_dst_filepath()
        self.fs.CreateDirectory(os.path.dirname(src))
        self.fs.CreateFile(src, contents=src_contents)
        self.fs.CreateDirectory(os.path.dirname(dst))
        self.fs.CreateFile(dst, contents=dst_contents)

    def test_rule_mismatch(self):
        task = io_scheduler.IO_Scheduler_Rule(mock.MagicMock())
        with self.assertRaises(case.Failure) as cm:
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self._setup_rule('1', '')
                task.run_test()
        self.assertEquals(cm.exception.message,
                          "contents of udev rule /etc/udev/rules.d/41-infinidat-io-scheduler.rules is not as expected")

    def test_rule_matches(self):
        task = io_scheduler.IO_Scheduler_Rule(mock.MagicMock())
        with mock.patch.object(io_scheduler, 'iter_infinidat_single_path_block_devices'):
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self._setup_rule('1', '1')
                task.run_test()

    def test_current_schedulers_not_noop(self):
        self.fs.CreateDirectory("/sys/block/sda/queue")
        self.fs.CreateFile("/sys/block/sda/queue/scheduler",
                           contents="noop [deadline] cfq")
        task = io_scheduler.IO_Scheduler_AffectedSettings(mock.MagicMock())
        with mock.patch.object(io_scheduler, 'iter_infinidat_single_path_block_devices') as _iter:
            _iter.return_value = ['/dev/sda']
            with self.assertRaises(case.Failure) as cm, mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self._setup_rule('1', '1')
                task.run_test()
        self.assertEquals(cm.exception.message,
                          "some INFINIDAT devices have the wrong I/O scheduler; for example, /dev/sda has deadline")

    def test_current_schedulers_noop(self):
        self.fs.CreateDirectory("/sys/block/sda/queue")
        self.fs.CreateFile("/sys/block/sda/queue/scheduler",
                           contents="[noop] deadline cfq")
        task = io_scheduler.IO_Scheduler_AffectedSettings(mock.MagicMock())
        with mock.patch.object(io_scheduler, 'iter_infinidat_single_path_block_devices') as _iter:
            _iter.return_value = ['/dev/sda']
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self._setup_rule('1', '1')
                task.run_test()


class FixTestCase(fake_filesystem_unittest.TestCase, FakeMixin):
    @classmethod
    def setUpClass(cls):
        if os.name == 'nt':
            raise SkipTest("this test case relies on posix file paths")

    def setUp(self):
        self.setUpPyfakefs()
        patch = mock.patch.object(io_scheduler, 'get_platform_string', new=lambda: 'linux-redhat-6')
        patch.__enter__()
        self.addCleanup(patch.__exit__, (None, None, None))

    def tearDown(self):
        self.tearDownPyfakefs()

    def test_fix_runtime(self):
        task = io_scheduler.IO_Scheduler_AffectedSettings(mock.MagicMock())
        self.fs.CreateDirectory("/sys/block/sda/queue")
        self.fs.CreateFile("/sys/block/sda/queue/scheduler",
                           contents="noop [deadline] cfq")
        with mock.patch.object(io_scheduler, 'iter_infinidat_single_path_block_devices') as _iter:
            _iter.return_value = ['/dev/sda']
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self._setup_rule('1', '')
                task.run_fix()

    def test_fix_rule(self):
        task = io_scheduler.IO_Scheduler_Rule(mock.MagicMock())
        self.fs.CreateDirectory("/sys/block/sda/queue")
        self.fs.CreateFile("/sys/block/sda/queue/scheduler",
                           contents="noop [deadline] cfq")
        with mock.patch.object(io_scheduler, 'iter_infinidat_single_path_block_devices') as _iter:
            _iter.return_value = ['/dev/sda']
            with mock.patch.object(io_scheduler, '_get_config', new=_get_fake_config):
                self._setup_rule('1', '')
                task.run_fix()
