from .....case import FixableCase

# pylint: disable=R0904,W0201

class SoftwareDependency(FixableCase):
    def __init__(self, category, kb_number, package_name, package_manager_class, specific_version=None):
        super(SoftwareDependency, self).__init__(category)
        self._kb_number = kb_number
        self._package_name = package_name
        self._package_manager_class = package_manager_class
        self._specific_version = specific_version

    def run_test(self):
        self._package_manager = self.get_package_manager()
        self.assertTrue(self._package_manager.is_package_installed(self.get_package_name()),
                        "package {} is not installed".format(self.get_package_name()))

    def get_fix_description(self):
        return "install missing package {} using the system package manager".format(self.get_package_name())

    def run_fix(self):
        # Backward compatibility
        if self._specific_version:
            self._package_manager.install_package(self.get_package_name(), self._specific_version)
        else:
            self._package_manager.install_package(self.get_package_name())

    def get_description(self):
        return "Checking that {} is installed".format(self.get_package_name())

    def get_kb_number(self):
        return self._kb_number

    def get_package_name(self):
        return self._package_name

    def get_package_manager(self):
        return self._package_manager_class()


class SoftwareVersionDependency(SoftwareDependency):
    def __init__(self, category, kb_number, package_name, package_version, package_manager_class):
        super(SoftwareVersionDependency, self).__init__(category, kb_number, package_name, package_manager_class)
        self._package_version = package_version

    def run_test(self):
        from infi.os_info.parse_version import parse_version
        super(SoftwareVersionDependency, self).run_test()

        installed_version = self._package_manager.get_installed_version(self.get_package_name())
        msg = "package {} version {} is lower than required {}"
        self.assertTrue(parse_version(installed_version['version']) > parse_version(self._package_version),
                        msg.format(self.get_package_name(), installed_version['version'], self._package_version))

    def get_fix_description(self):
        msg = "install package {} with version greater than {} using the system package manager"
        return msg.format(self.get_package_name(), self._package_version)

    def run_fix(self):
        if self._package_manager.is_package_installed(self.get_package_name()):
            self._package_manager.upgrade_package(self.get_package_name())
        else:
            self._package_manager.install_package(self.get_package_name())

    def get_description(self):
        msg = "Checking that {} is installed with version greater than {}"
        return msg.format(self.get_package_name(), self._package_version)
