
from ....case import FixableCase, Category
from infi.diskmanagement import MountManager, PartitionManager
from infi.diskmanagement.ioctl import constants
from infi.registry import LocalComputer, RegistryValueFactory

# pylint: disable-all

SAN_POLICIES = {
    "ONLINE":         constants.DISK_SAN_POLICY_ONLINE,
    "OFFLINE_SHARED": constants.DISK_SAN_POLICY_OFFLINE_SHARED,
    "OFFLINE":        constants.DISK_SAN_POLICY_OFFLINE
}


class SANPolicy(FixableCase):

    required_for_provisioning = True

    def run_test(self):
        policy = self._get_required_san_policy()
        self.assertTrue(PartitionManager().get_san_policy() == SAN_POLICIES[policy],
                        "SAN policy is not " + policy)

    def get_fix_description(self):
        return "set the SAN policy to OnlineAll using the Partition Manager APIs"

    def run_fix(self):
        PartitionManager().set_san_policy(SAN_POLICIES["ONLINE"])

    def get_description(self):
        policy = self._get_required_san_policy()
        return "Checking that the SAN policy is " + policy

    def get_kb_number(self):
        return "202319372"

    def _get_required_san_policy(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.disk.san_policy')


class AutoMounter(FixableCase):

    required_for_provisioning = True

    def run_test(self):
        self.assertFalse(MountManager().is_auto_mount(), "Auto mounter is enabled")

    def get_fix_description(self):
        return "disable auto-mount using Windows APIs"

    def run_fix(self):
        MountManager().disable_auto_mount()

    def get_description(self):
        return "Checking that volumes are not automatically mounted upon arrival"

    def get_kb_number(self):
        return "202404301"


DISK_PATH = r"System\CurrentControlSet\Services\Disk"

class DiskTimeout(FixableCase):
    def run_test(self):
        key = LocalComputer().local_machine[DISK_PATH]
        try:
            value = key.values_store['TimeOutValue'].to_python_object()
        except KeyError:
            value = 'undefined'
        recommended_timeout = self._get_recommended_timeout()
        self.assertEqual(value, recommended_timeout,
                         "Disk timeout value is {} instead of {}".format(value, recommended_timeout))

    def _get_recommended_timeout(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.disk.timeout')

    def get_fix_description(self):
        return """ change disk timeout value in the Registry.
        A reboot is required for this change to take effect """

    def run_fix(self):
        recommended_timeout = self._get_recommended_timeout()
        key = LocalComputer().local_machine[DISK_PATH]
        value = RegistryValueFactory().by_value(recommended_timeout)
        key.values_store['TimeOutValue'] = value
        self.ask_for_reboot("Timeout value has changed and a reboot is required for it to take effect")

    def get_description(self):
        recommended_timeout = self._get_recommended_timeout()
        return "Checking that the disk timeout value is {}".format(recommended_timeout)

    def get_kb_number(self):
        return "202404311"


def get_category():
    return Category("Disk Management", [SANPolicy, AutoMounter, DiskTimeout])
