
import os
from infi.wmpio import mpclaim
from infi.winver import Windows
from ....case import FixableCase, Category
from infi.wmpio.mpclaim import MultipathClaim
from infi.vendata.powertools.utils import is_veritas_multipathing_installed

# pylint: disable-all

LOAD_BALANCING_POLICIES = {
    "CLEAR_POLICY":            mpclaim.CLEAR_POLICY,
    "FAIL_OVER_ONLY":          mpclaim.FAIL_OVER_ONLY,
    "ROUND_ROBIN":             mpclaim.ROUND_ROBIN,
    "ROUND_ROBIN_WITH_SUBSET": mpclaim.ROUND_ROBIN_WITH_SUBSET,
    "LEAST_QUEUE_DEPTH":       mpclaim.LEAST_QUEUE_DEPTH,
    "WEIGHTED_PATHS":          mpclaim.WEIGHTED_PATHS,
    "LEAST_BLOCKS":            mpclaim.LEAST_BLOCKS,
}

class MultipathFeature(FixableCase):

    required_for_provisioning = True

    def run_test(self):
        from .feature import MultipathIoFeature
        self.assertTrue(MultipathIoFeature().is_installed(),
                        "Multipath I/O feature is not installed")

    def get_fix_description(self):
        return """ install the MPIO feature.
        A reboot is required for this change to take effect """

    def run_fix(self):
        from .feature import MultipathIoFeature, RebootRequired
        try:
            MultipathIoFeature().install()
        except RebootRequired:
            self.ask_for_reboot("Reboot is required for MultipathIo to start")

    def get_description(self):
        return "Checking that Multipath I/O feature is installed, this may take a while"

    def get_kb_number(self):
        return "202762962"

class ClaimingRules(FixableCase):

    required_for_provisioning = True

    def run_test(self):
        from infi.storagemodel.vendor.infinidat.infinibox import vid_pid_with_spaces
        vid, pid = vid_pid_with_spaces
        if not os.path.exists(MultipathClaim.path()):
            self.skip("Multipath I/O feature is not installed")
        self.assertTrue(MultipathClaim.is_hardware_claimed(vid, pid),
                        "INFINIDAT vendor and product identifiers are not claimed by MPIO")

    def get_fix_description(self):
        return """ claim NFINIDAT/InfiniBox devices using the mpclaim utility.
        A reboot is required for this change to take effect """

    def run_fix(self):
        from infi.storagemodel.vendor.infinidat.infinibox import vid_pid_with_spaces
        vid, pid = vid_pid_with_spaces
        MultipathClaim.claim_specific_hardware(vid, pid)

    def get_description(self):
        return "Checking Multipath claiming rules"

    def get_kb_number(self):
        return "202763012"

class DefaultLBPolicy(FixableCase):
    def run_test(self):
        from infi.storagemodel.vendor.infinidat.infinibox import vid_pid_with_spaces
        vid, pid = vid_pid_with_spaces
        if Windows().is_windows_2008():
            return
        if not os.path.exists(MultipathClaim.path()):
            self.skip("Multipath I/O feature is not installed")
        policy = MultipathClaim.get_hardware_specific_load_balancing_poicy(vid, pid)
        if policy == mpclaim.CLEAR_POLICY:
            policy = MultipathClaim.get_default_load_balancing_policy()
        expected = self._get_expected_value()
        self.assertEqual(policy, LOAD_BALANCING_POLICIES[expected],
                        "Default load balancing policy for InfiniBox devices is not " + expected)

    def get_fix_description(self):
        return "set the load balancing policy for InfiniBox to round robin using the mpclaim utility."

    def run_fix(self):
        from infi.storagemodel.vendor.infinidat.infinibox import vid_pid_with_spaces
        vid, pid = vid_pid_with_spaces
        if Windows().is_windows_2008():
            return
        expected = self._get_expected_value()
        MultipathClaim.set_hardware_specific_load_balancing_policy(vid, pid, LOAD_BALANCING_POLICIES[expected])

    def get_description(self):
        return "Checking multipath default load balancing policy for InfiniBox devices"

    def get_kb_number(self):
        return "202404281"

    def _get_expected_value(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.multipath.load_balancing_policy')


from infi.registry import LocalComputer, RegistryValueFactory
MPIO_PATH = r"SYSTEM\CurrentControlSet\Services\mpio"
MPIO_PARAMETERS_PATH = '\\'.join([MPIO_PATH, 'Parameters'])

class MPIORegistryKey(FixableCase):
    def run_test(self):
        expected = self._get_expected_value()
        try:
            key = LocalComputer().local_machine[MPIO_PARAMETERS_PATH]
            actual = key.values_store[self.value_name].to_python_object()
        except KeyError:
            self.fail("{} is not set to {}".format(self.value_name, expected))
        self.assertEqual(actual, expected,
                         "{} is set to {} instead of {}".format(self.value_name, actual, expected))

    def _get_expected_value(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.multipath')[self.value_name]

    def get_fix_description(self):
        return """ change registry value to the recommended value.
        A reboot is required for this change to take effect """

    def run_fix(self):
        try:
            LocalComputer().local_machine[MPIO_PARAMETERS_PATH]
        except:
            LocalComputer().local_machine[MPIO_PATH]['Parameters'] = None
        key = LocalComputer().local_machine[MPIO_PARAMETERS_PATH]
        expected = self._get_expected_value()
        key.values_store[self.value_name] = RegistryValueFactory().by_value(expected)
        msg = "{} value has changed and a reboot is required for it to take effect"
        self.ask_for_reboot(msg.format(self.value_name))

    def get_description(self):
        expected = self._get_expected_value()
        return "Checking that the {} value is {}".format(self.value_name, expected)

class PDORemovePeriod(MPIORegistryKey):
    value_name = 'PDORemovePeriod'

    def get_kb_number(self):
        return "202319362"

class UseCustomPathRecoveryInterval(MPIORegistryKey):
    value_name = 'UseCustomPathRecoveryInterval'

    def get_kb_number(self):
        return "202319362"

class PathRecoveryInterval(MPIORegistryKey):
    value_name = 'PathRecoveryInterval'

    def get_kb_number(self):
        return "202319362"

def get_category():
    if is_veritas_multipathing_installed():
        return Category("MPIO", [])
    else:
        return Category("MPIO", [MultipathFeature, ClaimingRules,
                                 PDORemovePeriod, UseCustomPathRecoveryInterval, PathRecoveryInterval])
