from __future__ import print_function

from ..case import Failure, Error, Warning, Skip, Reboot, CaseResultException
from infi.vendata.powertools.system_register.engine import confirm
from infi.vendata.powertools.utils.progress_bar import progress_bar_context
from infi.vendata.powertools.utils.file import safe_write_file, gevent_async_open
from infi.vendata.powertools import PROJECTROOT
import os
import sys
from logging import getLogger

logger = getLogger(__name__)

class ReadinessResult(object):
    def __init__(self):
        super(ReadinessResult, self).__init__()
        self.results = []
        self.count_by_result = {None: 0, Failure: 0, Error: 0, Reboot: 0, Warning: 0, Skip: 0}

    def add_result(self, case, exception):
        klass = None if exception is None else exception.__class__
        self.count_by_result[klass] += 1
        self.results.append((case, exception))

    def get_results(self):
        return self.results

    def is_host_ready(self):
        return self.count_by_result[Failure] + self.count_by_result[Error] + self.count_by_result[Reboot] == 0

    def is_reboot_required(self):
        return self.count_by_result[Reboot] > 0

    def is_provisioning_enabled(self):
        for case, exception in self.results:
            if case.required_for_provisioning and exception and exception.__class__ in (Failure, Error, Reboot):
                return False
        return True

    def get_counters(self):
        counters = {}
        for key, value in self.count_by_result.items():
            if key is None or value == 0:
                continue
            counters["{}s".format(key.__name__.lower())] = value
        return counters

    def to_json(self):
        from json import dumps
        cases = {}
        for case, exception in self.results:
            status = exception.__class__.__name__ if exception else 'OK'
            cases[case.__module__ + '.' + case.__class__.__name__] = status
        data = dict(cases=cases,
                    host_ready=self.is_host_ready(),
                    reboot_required=self.is_reboot_required(),
                    provisioning_enabled=self.is_provisioning_enabled())
        return dumps(data, indent=2, sort_keys=True)


KB_PREFIX = "http://support.infinidat.com/hc/articles"

def get_fore_color(name):
    from colorama import Fore
    return getattr(Fore, name)

class Status(object):
    _colors = {None:  get_fore_color("GREEN"),
               Failure: get_fore_color("RED"),
               Error: get_fore_color("MAGENTA"),
               Reboot: get_fore_color("CYAN"),
               Warning: get_fore_color("YELLOW"),
               Skip: get_fore_color("BLUE")}

    _messages = {None: 'ok',
                 Failure: 'fail',
                 Error: 'error',
                 Reboot: 'requires reboot',
                 Warning: 'warning',
                 Skip: 'skip'}

    _disable_print = False

    @classmethod
    def set_print_disabled(cls, disable_print):
        cls._disable_print = disable_print

    @classmethod
    def get_message(cls, result):
        klass = result.__class__ if result is not None else result
        msg = cls._messages[klass]
        color = cls._colors[klass]
        return msg, color

    @classmethod
    def print_status(cls, result, capital_case=False, new_line=False, stream=None):
        if stream is None:
            stream = sys.stdout
        if cls._disable_print:
            return
        msg, color = cls.get_message(result)
        if capital_case:
            msg = msg.capitalize()
        print(color + msg + get_fore_color("RESET"), end='\n' if new_line else ' ', file=stream)
        stream.flush()

    @classmethod
    def _get_fix_string(cls, case):
        import re
        fix_docstring = getattr(case, "get_fix_description")()
        if fix_docstring is None:
            return fix_docstring
        return re.sub("\n\s*", "\n" + " " * len("FIX STEPS: "), fix_docstring.strip())

    @classmethod
    def print_report(cls, case, result, print_qr_code=False, fix_message=False, stream=None):
        if stream is None:
            stream = sys.stdout
        if cls._disable_print:
            return
        klass = result.__class__ if result is not None else result
        color = cls._colors[klass]
        kb_url = "{}/{}".format(KB_PREFIX, case.get_kb_number())
        if not fix_message:
            cls.print_status(result, True, False, stream=stream)
            print(case, file=stream)
        print("REASON: {}".format(color + str(result) + get_fore_color("RESET")), file=stream)
        if fix_message:
            fix_str = cls._get_fix_string(case)
            if fix_str:
                print("FIX STEPS: {}".format(fix_str), file=stream)
        print("INFO: For more information, see {}".format(kb_url), file=stream)
        if print_qr_code:
            from infi.vendata.powertools.utils.qrcode import QRCodePool
            print(QRCodePool().get(kb_url), file=stream)

class SettingsRunner(object):
    def __init__(self, with_fix=False, with_auto_fix=False, with_qr_codes=False, disable_print=False):
        super(SettingsRunner, self).__init__()
        self.with_fix = with_fix
        self.with_auto_fix = with_auto_fix
        self.with_qr_codes = with_qr_codes
        self.disable_print = disable_print and not with_fix
        Status.set_print_disabled(self.disable_print)
        self.result = ReadinessResult()
        self.cases = self._build_case_list()
        self.some_failures_are_fixable = False

    def _get_platform_specific_case_module(self):
        from brownie.importing import import_string
        from infi.vendata.powertools.utils import get_platform_name
        platform_specific_module_name = '.'.join(["infi.vendata.powertools.settings.cases", get_platform_name()])
        try:
            logger.info("Importing platform-specific cases from {}".format(platform_specific_module_name))
            platform_specific_cases_module = import_string(platform_specific_module_name)
            return platform_specific_cases_module
            logger.info("Import successful")
        except ImportError:
            logger.exception("Import failed")
            raise

    def _build_case_list(self):
        cases = []
        for category in self.get_categories():
            for case in category.get_cases():
                cases.append(case(category))
        return cases

    def before_case(self, case, stream=None):
        if stream is None:
            stream = sys.stdout
        if not self.disable_print:
            print("{} ... ".format(case), end='', file=stream)
            stream.flush()

    def should_fix_this_case(self, case):
        raise NotImplementedError()

    def is_this_case_can_be_fixed(self, case):
        from ..case import FixableCase
        return isinstance(case, FixableCase)

    def handle_failure(self, case, result, stream=None):
        if stream is None:
            stream = sys.stdout
        # Check whether there's a fix, and whether to apply it
        if not self.is_this_case_can_be_fixed(case):
            logger.info("Case {!r} cannot be fixed".format(case))
            Status.print_status(result, capital_case=False, new_line=True, stream=stream)
            return result
        elif self.with_auto_fix:
            Status.print_status(result, capital_case=False, new_line=False, stream=stream)
        elif self.with_fix:
            Status.print_status(result, capital_case=False, new_line=True, stream=stream)
            Status.print_report(case, result, self.with_qr_codes, fix_message=True, stream=stream)
            if not confirm("Apply fix?"):
                self.some_failures_are_fixable = True
                return result
        else:
            logger.info("Case {!r} will not be fixed".format(case))
            self.some_failures_are_fixable = True
            Status.print_status(result, capital_case=False, new_line=False, stream=stream)
            print(get_fore_color("BLUE") + 'fix available ' + get_fore_color("RESET"), file=stream)
            stream.flush()
            return result
        # Apply fix
        try:
            if not self.disable_print:
                print(get_fore_color("BLUE") + 'applying fix ' + get_fore_color("RESET"), end='', file=stream)
                stream.flush()
            logger.info("Applying fix for case {!r}".format(case))
            self.execute_case_method(case, 'run_fix')
            logger.info("Testing again for case {!r}".format(case))
            self.execute_case_method(case, 'run_test')
            Status.print_status(None, False, True, stream=stream)
            return None
        except CaseResultException as result:
            logger.info("Case {!r} ended with: {!r}".format(case, result))
            Status.print_status(result, capital_case=False, new_line=True, stream=stream)
            if self.with_fix:
                print('-' * 79, file=stream)
                print('The last case failed during the fix, with the following error:')
                Status.print_report(case, result, self.with_qr_codes, stream=stream)
                print('-' * 79, file=stream)
            return result

    def execute_case_method(self, case, method_name='run_test'):
        from infi.storagemodel.errors import StorageModelError
        if not case.has_reboot_took_place():
            raise Reboot("A reboot is required from a previous run")
        try:
            msg = "Calling for method {} of {!r}"
            logger.debug(msg.format(method_name, case))
            result = getattr(case, method_name)()
            logger.debug("Method {} of {!r} returned {!r}".format(method_name, case, result))
        except CaseResultException as error:
            logger.debug("Method {} of {!r} raised {!r}".format(method_name, case, error))
            raise
        except StorageModelError as error:
            from traceback import format_exc
            msg = "An unhandled exception occurred while going over the storage stack. " + \
                  "Please rescan and try again; if the problem still occurs, please collect logs and contact support."
            logger.error(msg.format(method_name))
            logger.error(format_exc())
            case.error(error)
        except Exception as error:
            from traceback import format_exc
            msg = "An unhandled exception has been caught while calling for {}. " + \
                  "see the next log for exception and traceback"
            logger.error(msg.format(method_name))
            logger.error(format_exc())
            case.error(error)

    def run_case(self, case, stream=None):
        if stream is None:
            stream = sys.stdout
        try:
            with progress_bar_context(silent=self.disable_print):
                self.execute_case_method(case)
            self.result.add_result(case, None)
            Status.print_status(None, False, True, stream=stream)
            return None
        except Failure as result:
            logger.info("Case {!r} ended with: {!r}".format(case, result))
            result = self.handle_failure(case, result, stream=stream)
            self.result.add_result(case, result)
            return result
        except CaseResultException as result:
            logger.info("Case {!r} ended with: {!r}".format(case, result))
            self.result.add_result(case, result)
            Status.print_status(result, capital_case=False, new_line=True, stream=stream)
            return result

    def print_punch_lines(self, stream=None):
        if stream is None:
            self.stream = sys.stdout
        if self.result.is_host_ready():
            color, msg, ret = "GREEN", "This host is ready to work with the InfiniBox", 0
        elif self.result.is_reboot_required():
            color, msg, ret = "CYAN", "A reboot is required for changes to take effect", 1
        elif not self.with_fix and not self.with_auto_fix and self.some_failures_are_fixable:
            color, msg, ret = "YELLOW", "Re-run this tool with --fix flag to resolve issues that can be automatically fixed.", 2
        else:
            color, msg, ret = "MAGENTA", "This host is NOT ready to work with the InfiniBox", 3
        print(get_fore_color(color) + msg + get_fore_color("RESET"), file=stream)
        return ret

    def print_counters(self, stream=None):
        if stream is None:
            self.stream = sys.stdout
        counters = self.result.get_counters()
        if counters.keys():
            print("({})".format(', '.join(["{}={}".format(key, value) for key, value in counters.items()])), file=stream)

    def print_summary(self, stream=None):
        if stream is None:
            self.stream = sys.stdout
        if self.disable_print:
            return
        print('=' * 79, file=stream)
        for case, result in self.result.get_results():
            if result is not None:
                Status.print_report(case, result, self.with_qr_codes, stream=stream)
                print('-' * 79, file=stream)
        self.print_counters(stream=stream)
        return self.print_punch_lines(stream=stream)

    def save_result(self):
        """ Saves the readiness result in conf/readiness.json """
        path = ReadinessRunner.get_save_path()
        with safe_write_file(path) as fd:
            fd.write(self.result.to_json())

    @classmethod
    def load_previous_result(cls):
        """ Returns the parsed contents of the conf/readiness.json file, or None if it does not exist """
        from json import load
        path = ReadinessRunner.get_save_path()
        if not os.path.exists(path):
            return None
        with gevent_async_open(path, 'r') as fd:
            return load(fd)

    @classmethod
    def get_save_path(cls):
        return os.path.join(PROJECTROOT, 'conf', 'readiness.json')

    def run(self, stream=None):
        if stream is None:
            stream = sys.stdout
        logger.info("Running cases, with_fix={!r}, with_auto_fix={!r}, with_qr_codes={!r}".format(
                        self.with_fix, self.with_auto_fix, self.with_qr_codes))
        for case in self.cases:
            self.before_case(case, stream=stream)
            self.run_case(case, stream=stream)
        self.save_result()
        return self.print_summary(stream=stream)

class ReadinessRunner(SettingsRunner):
    def get_categories(self):
        return self._get_platform_specific_case_module().get_categories()


class DemoRunner(SettingsRunner):
    def get_categories(self):
        from ..cases.demo import get_category
        from ..cases import connectivity, devices
        return [get_category(), connectivity.get_category(), devices.get_category()]

class EmptyRunner(SettingsRunner):
    def get_categories(self):
        return []

class PassRunner(SettingsRunner):
    def get_categories(self):
        from ..cases.demo import get_category
        return [get_category(pass_only=True)]
