from __future__ import absolute_import

def get_platform_name():  # pragma: no cover
    from platform import system
    name = system().lower().replace('-', '_')
    return name

def get_distribution():  # pragma: no cover
    """:returns: munch with the following keys: distname, version, id
    """
    from munch import Munch
    from platform import linux_distribution
    distname, version, _id = linux_distribution()
    # distname in ['Red Hat Enterprise Linux Server', 'Ubuntu']
    distname = ''.join(distname.split()[:2]).lower()
    # This is a hack for centos (else it will be centoslinux)
    if distname.startswith('centos'):
        distname = 'centos'
    return Munch(distname=distname, version=version, id=_id.lower())

from infi.pyutils.lazy import wraps
import logging  # pylint: disable=W0403
logger = logging.getLogger(__name__)

WAIT_TIME = 120

def execute_command(cmd, check_returncode=True, timeout=WAIT_TIME):  # pragma: no cover
    from infi.execute import execute
    logger.info("executing {}".format(cmd))
    process = execute(cmd)
    process.wait(WAIT_TIME)
    logger.info("execution returned {}".format(process.get_returncode()))
    logger.debug("stdout: {}".format(process.get_stdout()))
    logger.debug("stderr: {}".format(process.get_stderr()))
    if check_returncode and process.get_returncode() != 0:
        formatted_cmd = cmd if isinstance(cmd, basestring) else repr(' '.join(cmd))
        raise RuntimeError("execution of {} failed".format(formatted_cmd))
    return process

def init_colors():
    from colorama import init
    from os import environ
    # see http://code.google.com/p/colorama/issues/detail?id=16
    # colors don't work on Cygwin if we call init
    # TODO delete this function when colorama is fixed
    if 'TERM' not in environ:  # this is how we recognize real Windows (init should only be called there)
        init()

def rescan_storagemodel(predicate, timeout_in_seconds=600):
    from infi.storagemodel import get_storage_model
    model = get_storage_model()
    model.rescan_and_wait_for(predicate, timeout_in_seconds=timeout_in_seconds)

def safe_get_device_partitions(device):
    from infi.storagemodel.base.disk import NoPartitionTable
    from infi.parted import InvalidPartitionTable, PartedNotInstalledException
    from infi.wioctl.errors import WindowsException
    try:
        return device.get_disk_drive().get_partition_table().get_partitions()
    except (InvalidPartitionTable, NoPartitionTable, IndexError, WindowsException, PartedNotInstalledException):
        return []

def get_mount_point(device):
    from infi.storagemodel import get_storage_model
    from infi.storagemodel.base.disk import NoSuchDisk
    try:
        disk_drive = device.get_disk_drive()
    except (NoSuchDisk, NotImplementedError):
        return None
    device_access_paths = disk_drive.get_block_access_paths_for_partitions()
    for mount in get_storage_model().get_mount_manager().get_mounts():
        if mount.get_block_access_path() in device_access_paths:
            return mount.get_mount_point()
    return None

def has_filesystem(device):
    return len(safe_get_device_partitions(device)) > 0

def is_mounted(device):
    return get_mount_point(device) is not None

def safe_rescan(func):
    """ Decorator that rescans and re-calls function if RescanIsNeeded is raised during execution """
    @wraps(func)
    def decorator(*args, **kwargs):
        from infi.storagemodel.errors import RescanIsNeeded
        from ..rescan.engine import do_rescan
        try:
            return func(*args, **kwargs)
        except RescanIsNeeded:
            do_rescan(disable_print=True)
            return func(*args, **kwargs)
    return decorator

def open_folder(path):
    """ Open a folder in the operating system's preferred viewer (e.g. Windows Explorer) """
    system = get_platform_name()
    if system == 'windows':
        import os
        os.startfile(path)
    elif system == 'darwin':
        execute_command(['open', path])
    else:
        execute_command(['xdg-open', path])

def round_to_block_size(bytes):
    """ Ensure that a given volume size is a round number of blocks """
    import math
    from infi.vendata.powertools.conf import get_configuration
    block_size = float(get_configuration().get('provisioning.block_size'))
    return int(math.ceil(bytes / block_size) * block_size)

def is_veritas_multipathing_installed():
    from os import path, environ
    veritas_executables = ('vxdmpadm',)
    return environ.get("VMPATH") or \
           any(path.exists(path.join(path.sep, "usr", "sbin", basename)) for basename in veritas_executables) or \
           bool(environ.get("MOCK_VERITAS"))

def get_infinidat_multipath_block_devices():
    from infi.storagemodel.vendor.infinidat import shortcuts
    if is_veritas_multipathing_installed():
        return shortcuts.get_infinidat_veritas_multipath_block_devices()
    else:
        return shortcuts.get_infinidat_native_multipath_block_devices()

def get_infinidat_non_multipath_block_devices():
    from infi.storagemodel.vendor.infinidat import shortcuts
    if is_veritas_multipathing_installed():
        return shortcuts.get_infinidat_non_veritas_multipathed_scsi_block_devices()
    else:
        return shortcuts.get_infinidat_non_multipathed_scsi_block_devices()
