import logging
import sys
import os
import time
import shutil
import docopt

from infi.pyutils.decorators import wraps
from infi.traceback import traceback_context
from infi.exceptools import exceptools_context

# this is here so that other modules in 'utils' can call 'from logging import ...'
from logging import getLogger, DEBUG   # noqa

# pylint: disable=W0201,C0103

def can_user_write_in_directory(dirpath):
    from os import path, remove, getpid
    src = path.join(dirpath, 'test-permissions.{}'.format(getpid()))
    try:
        with open(src, 'w'):
            pass
        remove(src)
    except IOError:
        return False
    return True

class LogFileHandler(logging.FileHandler):  # pragma: no cover
    def __init__(self, program_name, mode='a', encoding=None, delay=0):
        self._program_name = program_name
        logging.FileHandler.__init__(self, self.get_logfile_path(), mode=mode, encoding=encoding, delay=delay)

    def _get_logfile_filename(self):
        path = ".".join([self._program_name, self.get_timestamp(), "debug", "log"])
        index = 1
        while os.path.exists(os.path.join(self.get_logs_directory(), path)):
            path = ".".join([self._program_name, "{}-{}".format(self.get_timestamp(), index), "debug", "log"])
            index += 1
        return path

    def get_logfile_path(self):
        self._path = getattr(self, "_path",
                       os.path.join(self.get_logs_directory(), self._get_logfile_filename()))
        return self._path

    @classmethod
    def get_logs_directory(cls):
        if os.name == "nt":
            return os.environ.get("Temp", os.path.join(os.environ.get("SystemRoot", r"C:\Windows"), "Temp"))
        else:
            tmp = os.path.join(os.path.sep, 'tmp')
            var_log = os.path.join(os.path.sep, 'var', 'log')
            if can_user_write_in_directory(var_log):
                return var_log
            return tmp

    @classmethod
    def get_timestamp(self):
        return time.strftime("%Y-%m-%d.%H-%M")

root = logging.root

def get_executable_name():
    return os.path.basename(sys.argv[0])

POWERTOOLS_LOGGING_FORMATTER_KWARGS = dict(fmt='%(asctime)-25s %(levelname)-8s %(name)-50s %(message)s',
                                           datefmt='%Y-%m-%d %H:%M:%S %z')

class DebugLogging(object):  # pragma: no cover
    def __init__(self, program_name=None):
        object.__init__(self)
        self._program_name = program_name if program_name is not None else get_executable_name()

    def is_set(self):
        return any([isinstance(handler, LogFileHandler) for handler in root.handlers])

    def get_new_handler(self):
        handler = LogFileHandler(self._program_name)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(logging.Formatter(**POWERTOOLS_LOGGING_FORMATTER_KWARGS))
        return handler

    def set(self):
        root.addHandler(self.get_new_handler())

    def get(self):
        return filter(lambda handler: isinstance(handler, LogFileHandler), root.handlers)[0]

    def unset(self):
        root.removeHandler(self.get())

    def reset(self):
        for handler in list(root.handlers):
            root.removeHandler(handler)

    def safe_set(self):
        if not self.is_set():
            self.set()

    def safe_unset(self):
        if self.is_set():
            self.unset()

def print_general_error_message(msg):
    from infi.vendata.powertools.settings.runner import Status
    from infi.vendata.powertools.settings.case import Error
    from infi.vendata.powertools.settings.cases.general import GeneralErrorCase, get_category
    Status.print_report(GeneralErrorCase(get_category()), Error(msg), stream=sys.stderr)

def redirect_logbook():
    """
    Send logbook messages to the standard python logging (needed for infinisdk).
    """
    from logbook.compat import LoggingHandler
    from logbook.concurrency import enable_gevent
    LoggingHandler().push_application()
    enable_gevent()

class debug_program(object):  # pragma: no cover
    def __init__(self, program_name=None, print_path=False):
        self._program_name = program_name
        self._print_path = print_path
        self._log_path = None
        self._symlink = os.symlink if os.name != 'nt' else shutil.copy

    def __call__(self, func):
        """decorator for setting up a logging debug file
        Before calling the decorated function, a LogFileHandler is attached to the root logger,
        and removed after the function has exited (either returned or raised an exception).
        """
        @wraps(func)
        def callee(*args, **kwargs):
            from infi.storagemodel.errors import RescanIsNeeded, TimeoutError
            from infi.vendata.powertools.cli.engine import RescanFailed
            from infi.vendata.powertools.cli.validate import ValidationException
            from infi.vendata.powertools.utils.versioncheck import UnsupportedVersion
            from traceback import format_exception_only
            from sys import argv, exc_info
            redirect_logbook()
            logger = logging.getLogger(__name__)
            root.setLevel(logging.DEBUG)
            debug = DebugLogging(self._program_name)
            debug.safe_set()
            logger.info("Logging started")
            logger.info("Command-line arguments: {!r}".format(argv))
            handler = debug.get()
            self._log_path = handler.get_logfile_path()
            self._logdir = os.path.dirname(self._log_path)
            self._logname = ".".join([os.path.basename(self._log_path).split('.')[0], 'latest', 'debug', 'log'])
            with exceptools_context(), traceback_context():
                try:
                    return func(*args, **kwargs)
                except UnsupportedVersion as e:
                    print_general_error_message(e.message)
                    logger.exception(e.message)
                except docopt.DocoptExit as e:
                    from infi.vendata.powertools.cli import shorten_docopt_exit_string
                    logger.exception("DocoptExit raised")
                    shorten_docopt_exit_string(e)
                    raise
                except ValidationException as e:
                    logger.exception("ValidationException raised")
                    print e.args[0]
                except SystemExit:
                    pass
                except RescanFailed:
                    msg = "Some devices are not responding well. see log file {}".format(self._log_path)
                    print_general_error_message(msg)
                    logger.exception(msg)
                except (RescanIsNeeded, TimeoutError):
                    msg = "Some devices are missing; please rerun with '--rescan'. see log file {}".format(self._log_path)
                    print_general_error_message(msg)
                    logger.exception(msg)
                except:
                    formatted_exception = format_exception_only(*exc_info()[0:2])[0].strip()
                    msg = '{0}, see log file {1}'.format(formatted_exception, self._log_path)
                    print_general_error_message(msg)
                    logger.exception(msg)
                finally:
                    logger.info("Logging ended")
                    self._update_latest_symlink()
                    debug.safe_unset()
            return 1
        return callee

    def _remove_old_symbolic_link_if_exists(self, dst):
        if os.path.exists(dst):
            os.remove(dst)

    def _update_latest_symlink(self):
        src = self._log_path
        dst = os.path.join(self._logdir, self._logname)
        try:
            self._remove_old_symbolic_link_if_exists(dst)
            self._symlink(src, dst)
        except (OSError, IOError):
            pass

def suppress_warnings(func):
    @wraps(func)
    def callee(*args, **kwargs):
        import warnings
        with warnings.catch_warnings():
            warnings.resetwarnings()
            warnings.simplefilter("ignore")
            return func(*args, **kwargs)
    return callee

def log_usage(func):
    @wraps(func)
    def callee(*args, **kwargs):
        import platform
        from infi.vendata.powertools.__version__ import __version__
        log_name = get_executable_name() + ".usage.log"
        log_path = os.path.join(LogFileHandler.get_logs_directory(), log_name)
        timestamp = LogFileHandler.get_timestamp()
        msg = "{timestamp} PowerTools v{version} OS={os} {arch} argv={argv}\n"
        log_line = msg.format(argv=sys.argv,
                              version=__version__,
                              timestamp=timestamp,
                              os=platform.platform(),
                              arch=platform.machine())
        log_file = open(log_path, "a+")
        log_file.write(log_line)
        log_file.close()
        return func(*args, **kwargs)
    return callee
