from logging import getLogger
log = getLogger(__name__)


class ObjectNotFound(Exception):
    pass


class SystemNotFound(ObjectNotFound):
    pass


def get_infinibox_scsi_devices_mapped_to_lun0():
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
    return get_infinidat_block_devices_and_controllers__mapped_to_lun0()

def get_uniq_list_of_attached_boxes__by_serial():
    """:returns: a unique list of attached InfiniBoxes"""
    infiboxes = get_infinibox_scsi_devices_mapped_to_lun0()
    boxes_by_serial_number = {}
    for box in infiboxes:
        serial = box.get_vendor().get_system_serial()
        boxes_by_serial_number[serial] = boxes_by_serial_number.get(serial, box)
    return boxes_by_serial_number.values()

def get_connections_by_serial():
    """:returns: a {serial: (vendor, connections)} dictionary"""
    from infi.storagemodel.base.scsi import SCSIDevice
    from infi.storagemodel.base.multipath import MultipathDevice
    from infi.storagemodel.connectivity import FCConnectivity
    devices = get_infinibox_scsi_devices_mapped_to_lun0()
    boxes = {}
    for device in devices:
        serial = device.get_vendor().get_system_serial()
        if serial not in boxes:
            boxes[serial] = device.get_vendor(), []
        connections = [device.get_connectivity()] if isinstance(device, SCSIDevice) else \
                      [path.get_connectivity() for path in device.get_paths()] if isinstance(device, MultipathDevice) else \
                      []
        valid_connections = [connection for connection in connections if isinstance(connection, FCConnectivity)]
        boxes[serial][-1].extend(valid_connections)
    return boxes

def get_uniq_list_of_attached_boxes__by_initiator_and_serial():
    """:returns: a dict of (serial, initiator_wwn): controller of attached InfiniBoxes"""
    infiboxes = get_infinibox_scsi_devices_mapped_to_lun0()
    boxes = {}
    for box in infiboxes:
        serial = box.get_scsi_serial_number()
        initiator = get_initiator_wwn_from_controller_device(box)
        if initiator is None:
            log.error("did not find initiator wwn for InfiniBox device {!r}, this is a bug".format(box))
        else:
            boxes[(serial, initiator)] = boxes.get((serial, initiator), box)
    return boxes.values()

def get_host_id_from_controller(controller):
    host_id = controller.get_vendor().get_host_id()
    return host_id

def is_host_port_defined(controller):
    is_defined = get_host_id_from_controller(controller) != 0
    return is_defined

def get_system_name_from_controller(controller):
    return controller.get_vendor().get_system_name()

def get_system_serial_from_controller(controller):
    return controller.get_vendor().get_system_serial()

def get_defined_host_object(system_object, controller):
    host_id = get_host_id_from_controller(controller)
    return get_host_object_by_id(system_object, host_id)

def get_initiator_wwn_from_controller_device(controller):
    from infi.storagemodel.connectivity import FCConnectivity
    if isinstance(controller.get_connectivity(), FCConnectivity):
        return controller.get_connectivity().get_initiator_wwn()

def get_target_wwn_from_controller_device(controller):
    return controller.get_connectivity().get_target_wwn()

def are_there_no_ports_on_host(host_object):
    return len(host_object.get_fc_ports()) == 0

def is_host_object_exists(system_object, host_name):
    result = any(system_object.hosts.find(name=host_name))
    log.info("{} with name {!r} on system {!r}".format("Found host" if result else "There is no such host",
                                             host_name, system_object.get_name()))
    return result

def get_host_id_by_initiator_address(system_object, wwn):
    host = system_object.hosts.get_host_by_initiator_address(wwn)
    return host.get_id() if host else None

def is_any_of_our_ports_registered_in_system(system_object, our_ports):
    host_ids = {get_host_id_by_initiator_address(system_object, wwn) for wwn in our_ports}
    host_ids.discard(None)
    return len(host_ids) > 0

def are_all_ports_belong_to_the_same_host(system_object, our_ports):
    host_ids = {get_host_id_by_initiator_address(system_object, wwn) for wwn in our_ports}
    return len(host_ids) == 1 and None not in host_ids and \
           get_host_object_by_id(system_object, host_ids.pop()).get_fc_ports() == our_ports

def get_host_objects_by_our_ports(system_object, our_ports):
    host_ids = {get_host_id_by_initiator_address(system_object, wwn) for wwn in our_ports}
    return [get_host_object_by_id(system_object, host_id) for host_id in host_ids]

def get_host_object_by_name(system_object, host_name):
    result = system_object.hosts.safe_get(name=host_name)
    if result is None:
        raise ObjectNotFound("Could not find host with name {}".format(host_name))
    log.info("Found in system {!r} a host with id {!r} by name {!r}".format(system_object.get_name(),
                                                                            result.get_id(), host_name))
    return result

def get_host_object_by_id(system_object, host_id):
    result = system_object.hosts.safe_get(id=host_id)
    if result is None:
        raise ObjectNotFound("Could not find host with id {}".format(host_id))
    log.info("Found in system {!r} a host by id {!r} with name {!r}".format(system_object.get_name(),
                                                                            host_id, result.get_name()))
    return result

def get_host_object_by_wwn(system_object, host_wwn):
    host_id = get_host_id_by_initiator_address(system_object, host_wwn)
    return None if host_id is None else get_host_object_by_id(system_object, host_id)

def get_foreign_ports_from_host(host_object, our_ports):
    log.info("host with id={!r} has the following fc ports: {!r}".format(host_object.get_id(),
                                                                         host_object.get_fc_ports()))
    log.info("our ports are: {!r}".format(our_ports))
    return [port for port in host_object.get_fc_ports() if port not in our_ports]

def host_object_has_other_ports(host_object, our_ports):
    result = any(get_foreign_ports_from_host(host_object, our_ports))
    log.info("thus the host {} other ports".format("indeed has" if result else "does not have"))
    return result

def are_all_ports_match(host_object, our_ports):
    return set(our_ports).issuperset(set(host_object.get_fc_ports()))

def get_serial_from_controller(controller):
    return controller.get_scsi_serial_number()

def are_there_foreign_ports_on_host(host_object, our_ports):
    return host_object_has_other_ports(host_object, our_ports)

def is_same_device(serial, initiator, target, other_device):
    return serial == other_device.get_scsi_serial_number() and \
        initiator == other_device.get_connectivity().get_initiator_wwn() and \
        target == other_device.get_connectivity().get_target_wwn()

def get_os_hostname():
    from socket import getfqdn
    return getfqdn()

def get_os_platform():
    from platform import platform
    from infi.os_info import get_platform_string
    result = platform()
    if result.lower().startswith("aix"):   # HPT-1695
        result = get_platform_string()
    return result

def get_os_system():
    from platform import system
    return system()

def get_powertools_version():
    from infi.vendata.powertools.__version__ import __version__
    return __version__

def get_system_object(serial, address, port, log_in=True):
    """:returns: infinisdk.InfiniBox object"""
    from infinisdk.core.exceptions import APITransportFailure
    from .. import auth
    from infi.vendata.infinibox_credentials_store.utils import get_cached_system
    if log_in:
        credentials = auth.CREDENTIALS_STORE.get_credentials(serial)
        system = get_cached_system(address, credentials.get_username(), credentials.get_password())
    else:
        system = get_cached_system(address, None, None)
    system.api.add_auto_retry(lambda e: isinstance(e, APITransportFailure) and
                                        "Interrupted system call" in e.error_desc, 5)
    system.api.set_source_identifier("PowerTools/{}".format(get_powertools_version()))
    if log_in:
        # simulate session by calling infinibox login api. infinisdk will take the credentials we already set
        system.login()
    return system

def get_system_controller_by_serial(system_serial):
    for controller in get_infinibox_scsi_devices_mapped_to_lun0():
        serial = get_system_serial_from_controller(controller)
        if serial == int(system_serial):
            return controller
    raise SystemNotFound("Could not find system {}".format(system_serial))

def get_system_object_by_device_vendor(vendor, log_in=True):
    _validate_version(vendor)
    serial, address, port = vendor.get_system_serial(), vendor.get_management_address(), vendor.get_management_port()
    system = get_system_object(serial, address, port, log_in)
    return system

def get_system_object_by_device(device, log_in=True):
    return get_system_object_by_device_vendor(device.get_vendor(), log_in)

def get_system_object_by_serial(system_serial, log_in=True):
    controller = get_system_controller_by_serial(system_serial)
    system = get_system_object_by_device(controller, log_in)
    return system

def get_pools(system_object):
    """:returns: list of Pool objects that the current user can administer"""
    return list(system_object.pools.get_administered_pools())

def get_system_object_and_host(system_serial, log_in=True):
    controller = get_system_controller_by_serial(system_serial)
    system = get_system_object_by_device(controller, log_in)
    host = get_defined_host_object(system, controller)
    return system, host

def get_infinisdk_host_and_volume_by_device(infinidat_device):
    system = get_system_object_by_device(infinidat_device)
    vendor = infinidat_device.get_vendor()
    volume = system.volumes.safe_get(id=vendor.get_volume_id())
    if volume is None:
        raise ObjectNotFound("Could not find volume with id {}".format(vendor.get_volume_id()))
    host = system.hosts.safe_get(id=vendor.get_host_id())
    if host is None:
        raise ObjectNotFound("Could not find host with id {}".format(vendor.get_host_id()))
    return host, volume

def _validate_version(vendor):
    from infi.vendata.powertools.utils.versioncheck import raise_if_unsupported
    raise_if_unsupported(vendor.get_system_version())

def is_device_read_only(device):
    try:
        return device.get_vendor().check_if_write_protected()
    except:
        log.exception("SCSI check to see if device is write protected failed, switching to API")
        host, volume = get_infinisdk_host_and_volume_by_device(device)
        return volume.is_snapshot() or volume.get_write_protected()
