import operator


class Column(object):

    def __init__(self, key, label=None, formatter=None, right_aligned=False):
        """
        Defines one table column.
        key - either a name or an index, depending on whether each row is a mapping or a sequence
        label - the column label
        formatter - a function for converting the row value into a string
        right_aligned - pass True for aligning the values to the right
        """
        self.key = key
        self.is_name = isinstance(self.key, basestring)
        self.label = label or str(key).replace('_', ' ').title()
        self.formatter = formatter or str
        self.right_aligned = right_aligned

    def layout(self, value, width):
        """ Layout the given string into the requested fixed width """
        if len(value) > width:
            value = value[:width - 3] + '...'
        padding = ' ' * (width - len(value))
        return padding + value if self.right_aligned else value + padding

    def widest(self, rows):
        """ Find the maximum width needed for this column """
        width = max((len(self.get_value(row)) for row in rows)) if rows else 0
        return max(width, len(self.label))

    def get_value(self, row):
        """ Get the formatted value for this column from the given row """
        value = row.getattr(self.key) if self.is_name else row[self.key]
        return '-' if value is None else self.formatter(value)


class Table(object):
    """
    A class for printing tabular data. The width of each column are determined by
    its contents, and clipped if necessary to fit into the maximum allowed width.
    """
    def __init__(self, columns, max_width=250):
        self.columns = columns
        self.max_width = max_width
        self.rows = []

    def add_row(self, row):
        self.rows.append(row)

    def __str__(self):
        lines = []
        widths = self._calculate_widths()
        line = '  '.join([col.layout(col.label, widths[i]) for i, col in enumerate(self.columns)])
        lines.append(line)
        for row in self.rows:
            line = '  '.join([col.layout(col.get_value(row), widths[i]) for i, col in enumerate(self.columns)])
            lines.append(line)
        return '\n'.join(lines)

    def _calculate_widths(self):
        widths = [col.widest(self.rows) for col in self.columns]
        total_width = sum(widths) + 2 * len(widths)
        full_widths = list(widths)
        while total_width > self.max_width:
            # Choose which column to shorten so that the percentage of hidden characters in it will be the lowest
            hidden_ratios = []
            for i in range(len(widths)):
                chars = 3 if widths[i] == full_widths[i] else 1  # three is for the ellipsis
                hidden_ratios.append((full_widths[i] - widths[i] + chars) / float(full_widths[i]))
            index, ratio = min(enumerate(hidden_ratios), key=operator.itemgetter(1))
            widths[index] -= 1
            total_width -= 1
        # Add 2 characters for column spacing
        widths = [width + 2 for width in widths]
        return widths
