import sys

from infi.storagemodel.vendor.infinidat.infinibox.wwn import InfinidatWWN
from infi.storagemodel.errors import RescanIsNeeded
from infi.pyutils.lazy import cached_method
from infi.vendata.powertools.utils import is_veritas_multipathing_installed
from ...utils.progress_bar import progress_bar_context
from ...utils.termsize import getTerminalSize
from ...utils import get_infinidat_non_multipath_block_devices, get_infinidat_multipath_block_devices
import operator
from logging import getLogger
logger = getLogger(__name__)

# pylint: disable=R0922


class Connection(object):  # pragma: no cover
    def get_type(self):
        raise NotImplementedError()

    def get_initiator(self):
        raise NotImplementedError()

    def get_target(self):
        raise NotImplementedError()

    def get_access_path(self):
        raise NotImplementedError()

    def get_state(self):
        raise NotImplementedError()


class StorageModelFiberChannelConnection(Connection):
    def __init__(self, connectivity, access_path=None, state="Up"):
        super(StorageModelFiberChannelConnection, self).__init__()
        self._connectivity = connectivity
        self._access_path = access_path
        self._state = state

    def get_type(self):
        return "FC"

    def get_initiator(self):
        try:
            return str(self._connectivity.get_initiator_wwn())
        except:
            logger.debug("faild to get initiator wwn", exc_info=True)
            return 'N/A'

    def get_target(self):
        try:
            return str(self._connectivity.get_target_wwn())
        except:
            logger.debug("faild to get initiator wwn", exc_info=True)
            return 'N/A'

    def get_access_path(self):
        return self._access_path

    def get_state(self):
        return self._state

    def get_node_id(self):
        from infi.storagemodel.vendor.infinidat.infinibox.wwn import InfinidatWWN
        try:
            return InfinidatWWN(self.get_target()).get_node_id()
        except KeyError:
            return -1

    def get_port_id(self):
        try:
            return InfinidatWWN(self.get_target()).get_port_id()
        except KeyError:
            return -1

    def __repr__(self):
        msg = "<Connection(type={!r}, initiator={!r}, target={!r}, access_path={!r}, state={!r})>"
        return msg.format(self.get_type(), self.get_initiator(), self.get_target(), self.get_access_path(),
                          self.get_state())


class UnknownFiberChannelConnectivity(object):
    def get_initiator_wwn(self):
        return "N/A"

    def get_target_wwn(self):
        return "N/A"


class UnknownStorageModelFiberChannelConnection(StorageModelFiberChannelConnection):
    def __init__(self, access_path):
        super(UnknownStorageModelFiberChannelConnection, self).__init__(UnknownFiberChannelConnectivity(), access_path)

    def get_node_id(self):
        return -1

    def get_port_id(self):
        return -1


class Device(object):  # pragma: no cover
    def __lt__(self, other):
        return self.get_lun() < other.get_lun()

    def get_lun(self):
        raise NotImplementedError()

    def get_type(self):
        raise NotImplementedError()

    def get_scsi_serial(self):
        raise NotImplementedError()

    def get_volume_id(self):
        raise NotImplementedError()

    def get_volume_name(self):
        raise NotImplementedError()

    def get_size_in_bytes(self):
        raise NotImplementedError()

    def is_multipath(self):
        raise NotImplementedError()

    def get_multipath_framework_name(self):
        raise NotImplementedError()

    def get_number_of_paths(self):
        raise NotImplementedError()

    def get_access_path(self):
        raise NotImplementedError()

    def get_connections(self):
        raise NotImplementedError()

    def get_connection_count(self):
        return len(self.get_connections())

class System(object):
    def __init__(self, serial, name, management, devices):
        self._name = name
        self._serial = serial
        self._devices = devices
        self._management = management

    def get_serial(self):
        return self._serial

    def get_name(self):
        return self._name

    def get_management(self):
        return self._management

    def get_devices(self):
        """:returns: a list of :class:`.Device` objects, sorted by increasing LUN number"""
        return sorted(self._devices)

    def get_display_name(self):
        raise NotImplementedError()

    def __repr__(self):
        msg = "<System(serial={!r}, name={!r}, management={!r}, devices={!r}>"
        return msg.format(self._serial, self._name, self._management, self._devices)


class StorageModelDevice(Device):
    def __init__(self, device):
        super(StorageModelDevice, self).__init__()
        self._device = device

    def get_lun(self):
        from infi.storagemodel.base.scsi import SCSIDevice
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        if isinstance(self._device, SCSIDevice):
            try:
                return self._device.get_hctl().get_lun()
            except:
                logger.debug("failed to get lun from hctl", exc_info=True)
                return -1
        if isinstance(self._device, MultipathBlockDevice) or isinstance(self._device, MultipathStorageController):
            try:
                return self._device.get_paths()[0].get_hctl().get_lun()
            except:
                logger.debug("failed to get lun from hctl", exc_info=True)
                return -1
        msg = "{!r} is not an instance of SCSIDevice or MultipathBlockDevice or MultipathStorageController"  # pragma: no cover
        raise TypeError(msg.format(self._device))  # pragma: no cover

    def get_type(self):
        from infi.storagemodel.base.scsi import SCSIStorageController
        from infi.storagemodel.base.multipath import MultipathStorageController
        return "Controller" if isinstance(self._device, (SCSIStorageController, MultipathStorageController)) else "Volume"

    def get_scsi_serial(self):
        return self._device.get_scsi_serial_number()

    def get_volume_id(self):
        if self.get_type() == "Controller":
            return None
        return hex(self._device.get_vendor().get_naa().get_volume_id())

    def get_volume_name(self):
        if self.get_type() == "Controller":
            return None
        return self._device.get_vendor().get_volume_name() or "N/A"

    def get_size_in_bytes(self):
        from infi.storagemodel.base.scsi import SCSIBlockDevice, SCSIStorageController
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        if isinstance(self._device, (SCSIBlockDevice, MultipathBlockDevice)):
            return self._device.get_size_in_bytes()
        if isinstance(self._device, SCSIStorageController) or isinstance(self._device, MultipathStorageController):
            return None
        msg = "{!r} is not an instance of SCSIDevice or MultipathBlockDevice or SCSIStorageController or MultipathStorageController"  # pragma: no cover
        raise TypeError(msg.format(self._device))  # pragma: no cover

    def is_multipath(self):
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        return isinstance(self._device, MultipathBlockDevice) or isinstance(self._device, MultipathStorageController)

    def get_multipath_framework_name(self):
        from infi.storagemodel import get_storage_model
        from infi.os_info import get_platform_string
        if self._device in get_storage_model().get_native_multipath().get_all_multipath_block_devices():
            return "Native"
        elif is_veritas_multipathing_installed() and \
             self._device in get_storage_model().get_veritas_multipath().get_all_multipath_block_devices():
            if "windows" in get_platform_string():
                return "Native"  # in windows, veritas multipath is just a proxy in our code to the native multipath
            else:
                return "Veritas"
        return None

    def get_access_path(self):
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        from infi.storagemodel.base.scsi import SCSIDevice
        if isinstance(self._device, SCSIDevice):
            return self._device.get_scsi_access_path()
        elif isinstance(self._device, MultipathBlockDevice):
            return self._device.get_block_access_path()
        elif isinstance(self._device, MultipathStorageController):
            return self._device.get_multipath_access_path()
        else:
            return 'None'

    @cached_method
    def _get_list_of_matching_scsi_devices(self, serial_number):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_scsi_devices
        return filter(lambda device: device.get_scsi_serial_number() == serial_number,
                      get_infinidat_scsi_devices())

    def _get_block_access_path_for_path_object(self, path):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_scsi_block_devices
        all_scsi_devices = get_infinidat_scsi_block_devices()
        try:
            for device in all_scsi_devices:
                if device.get_hctl() == path.get_hctl():
                    return device.get_block_access_path()
        except:
            logger.debug("failed to get block access path for path object".format(path), exc_info=True)
        return '<unknown>'

    def get_mount_point(self):
        from ...utils import get_mount_point
        from infi.storagemodel.base.multipath import MultipathBlockDevice
        from infi.vendata.powertools.cli.engine import is_filesystem_support_disabled
        if is_filesystem_support_disabled():
            return None
        if not isinstance(self._device, MultipathBlockDevice):
            return None
        return get_mount_point(self._device)

    def get_display_name(self):
        from infi.storagemodel.base.scsi import SCSIStorageController
        if isinstance(self._device, SCSIStorageController):
            return None
        return self._device.get_display_name()

    def get_connections(self):
        from infi.storagemodel.base.scsi import SCSIDevice
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        from infi.storagemodel.connectivity import FCConnectivity
        if isinstance(self._device, SCSIDevice):
            return [StorageModelFiberChannelConnection(device.get_connectivity(), device.get_scsi_access_path())
                    if isinstance(device.get_connectivity(), FCConnectivity)
                    else UnknownStorageModelFiberChannelConnection(device.get_scsi_access_path())
                    for device in self._get_list_of_matching_scsi_devices(self._device.get_scsi_serial_number())]
        if isinstance(self._device, MultipathBlockDevice) or isinstance(self._device, MultipathStorageController):
            _get_path = self._get_block_access_path_for_path_object
            return [StorageModelFiberChannelConnection(path.get_connectivity(), _get_path(path),
                                                       path.get_state().capitalize())
                    if isinstance(path.get_connectivity(), FCConnectivity)
                    else UnknownStorageModelFiberChannelConnection(_get_path(path))
                    for path in self._device.get_paths()]

    def get_connection_count(self):
        from infi.storagemodel.base.scsi import SCSIDevice
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        if isinstance(self._device, SCSIDevice):
            return len(self._get_list_of_matching_scsi_devices(self._device.get_scsi_serial_number()))
        if isinstance(self._device, MultipathBlockDevice) or isinstance(self._device, MultipathStorageController):
            return len(self._device.get_paths())

    def __repr__(self):
        msg = "<StorageModelDevice(lun={!r}, type={!r}, scsi_serial={!r}, volume_id={!r}, " + \
                "volume_name={!r}, size_in_bytes={}, is_multipath={!r}, display_name={!r}, connections={!r})>"
        return msg.format(self.get_lun(), self.get_type(), self.get_scsi_serial(), self.get_volume_id(),
                          self.get_volume_name(), self.get_size_in_bytes(), self.is_multipath(),
                          self.get_display_name(), self.get_connections())


class Inventory(object):
    def __init__(self):
        super(Inventory, self).__init__()
        self._systems = []
        self._exc_info = None

    def populate(self):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_storage_controller_devices
        devices = ([], [], [])
        try:
            devices = (get_infinidat_storage_controller_devices(),
                       get_infinidat_multipath_block_devices(),
                       get_infinidat_non_multipath_block_devices())
        except:
            logger.exception("Failed to detect InfiniBox devices")
            self._exc_info = sys.exc_info()
        try:
            self._group_by_system(*devices)
        except:
            logger.exception("Failed to detect InfiniBox devices")
            self._exc_info = sys.exc_info()

    def _iter_uniq_serial_number(self, items, serial_numbers=None):
        serial_numbers = serial_numbers if serial_numbers else set()
        for item in items:
            try:
                serial_number = item.get_scsi_serial_number()
            except RescanIsNeeded:
                logger.error("failed to get scsi serial for {}".format(item))
                continue
            if serial_number in serial_numbers:
                continue
            serial_numbers.add(serial_number)
            msg = "Found a leading device {!r} for devices serial number {}"
            logger.debug(msg.format(item, serial_number))
            yield item

    def _group_by_system(self, controllers, multipath_volumes, non_multipath_volumes):
        by_system_serial = {}
        from infi.gevent_utils.silent_greenlets import spawn, joinall
        def thread_func(device):
            serial = device.get_vendor().get_system_serial()
            devices = by_system_serial.setdefault(serial, list())
            devices.append(device)
        greenlets = []
        for device in (list(self._iter_uniq_serial_number(controllers)) +
                       multipath_volumes +
                       list(self._iter_uniq_serial_number(non_multipath_volumes))):
            greenlets.append(spawn(thread_func, device))
        joinall(greenlets)
        for system_serial, devices in by_system_serial.items():
            try:
                management = devices[0].get_vendor().get_management_address()
                system_name = devices[0].get_vendor().get_system_name()
            except:
                logger.error("failed to get InfiniBox management address or name from scsi inquiry of {}".format(devices[0]))
                management, system_name = '<unknown>', '<unknown>'
            uniq_devices = [StorageModelDevice(item) for item in devices]
            system = System(str(system_serial), system_name, management, uniq_devices)
            self._systems.append(system)

    def get_systems(self):
        return [item for item in self._systems]


class OutputStrategy(object):
    def execute(self, inventory):  # pragma: no cover
        raise NotImplementedError()

    def set_size_units(self, size_units):
        self.size_units = size_units

    def _strip_line(self, line, maximum_length):
        line_str = str(line) if not isinstance(line, (str, unicode,)) else line
        if len(line_str) <= maximum_length:
            return line_str
        return line_str[:maximum_length - 3] + '...'

    def _format(self, line, length):
        return self._strip_line(line, length - 2).ljust(length)

    def _size_to_string(self, size_in_bytes):
        from capacity import byte
        if size_in_bytes is None:
            return '-'
        return str(byte * size_in_bytes).replace('*', ' ')


class BriefTerminalOutput(OutputStrategy):
    COLUMN_NAMES = ["Name", "Device", "Mount", "Size", "# Paths", "LUN"]

    def _size_to_string(self, size_in_bytes):
        import capacity
        padding = self._column_sizes[self.COLUMN_NAMES.index("Size")] - 3
        size_unit = getattr(capacity, self.size_units)
        if size_in_bytes is None:
            return '-'.rjust(padding)
        size_unit_in_bytes = size_unit / capacity.byte
        return "{:.2f}".format(float(size_in_bytes) / float(size_unit_in_bytes)).rjust(padding)

    def execute(self, inventory):
        self._inventory = inventory  # pylint: disable=W0201
        systems = self._inventory.get_systems()
        if not systems:
            print "No InfiniBoxes detected          "
            return
        self._column_sizes = self._calculate_column_sizes(systems)
        for system in systems:
            self._print_system(system)
            print ""
        print "* Non-multipathed"
        print ""

    def _calculate_column_sizes(self, systems):
        # Find the required size of each column
        from infi.gevent_utils.silent_greenlets import spawn, joinall
        fixed_column_sizes = [11, 10, 8]
        sizes = [12, 12, 12]
        for system in systems:
            greenlets = []
            volume_names = []
            display_names = []
            mount_points = []
            for device in system.get_devices():
                greenlets.append(spawn(lambda device: volume_names.append(device.get_volume_name()), device))
                greenlets.append(spawn(lambda device: display_names.append(device.get_display_name()), device))
                greenlets.append(spawn(lambda device: mount_points.append(device.get_mount_point()), device))
            joinall(greenlets)
            sizes[0] = max(sizes[0], *[len(i or '') for i in volume_names])
            sizes[1] = max(sizes[1], *[len(i or '') for i in display_names])
            sizes[2] = max(sizes[2], *[len(i or '') for i in mount_points])
        # If the result is too wide, reduce the sizes to fit
        max_size = getTerminalSize()[0] - 1
        total_size = sum(sizes) + sum(fixed_column_sizes) + 6  # six is for column spacing
        full_sizes = list(sizes)
        while total_size > max_size:
            # Choose which column to shorten so that the percentage of hidden characters in it will be the lowest
            hidden_ratios = []
            for i in range(3):
                chars = 3 if sizes[i] == full_sizes[i] else 1  # three is for the ellipsis
                hidden_ratios.append((full_sizes[i] - sizes[i] + chars) / float(full_sizes[i]))
            index, ratio = min(enumerate(hidden_ratios), key=operator.itemgetter(1))
            sizes[index] -= 1
            total_size -= 1
        # Add 2 characters for column spacing
        sizes = [size + 2 for size in sizes]
        # Append the remaining fixed-size columns and return the result
        return sizes + fixed_column_sizes

    def _print_system_header(self, system):
        header = "System (name=, serial={}):".format(system.get_serial())
        remaining = getTerminalSize()[0] - len(header) - 1
        header = header.replace('name=', 'name=%s' % self._strip_line(system.get_name(), remaining), 1)
        print header

    def _print_column_names(self, system):
        import capacity
        size_unit = getattr(capacity, self.size_units)
        names = self.COLUMN_NAMES[:]
        names[names.index("Size")] = "Size({})".format(str(size_unit).replace("1*", ""))
        line = ''.join([name.ljust(size) for name, size in zip(names, self._column_sizes)])
        print line

    def _print_device(self, device):
        size_str = self._size_to_string(device.get_size_in_bytes())
        path = device.get_display_name()
        path_str = '-' if path is None else path
        numdevs_str = "{}{}".format(str(device.get_connection_count()),
                                    "*" if not device.is_multipath() else "")
        name = device.get_volume_name()
        mount_point = device.get_mount_point()
        mount_point_str = '-' if mount_point is None else mount_point
        name_str = '(Controller)' if name is None else name
        values = [name_str, path_str, mount_point_str, size_str,
                  numdevs_str, device.get_lun()]
        line = ''.join([self._format(value, size) for value, size in zip(values, self._column_sizes)])
        print line

    def _print_system(self, system):
        self._print_system_header(system)
        self._print_column_names(system)
        for device in system.get_devices():
            self._print_device(device)


class DetailedTerminalOutput(OutputStrategy):
    KEY_LENGTH = 35

    def execute(self, inventory):
        self._inventory = inventory  # pylint: disable=W0201
        self._indent = 0  # pylint: disable=W0201
        systems = self._inventory.get_systems()
        if not systems:
            print "No InfiniBoxes detected"
            return
        for system in systems:
            self._print_system(system)
            print ""

    def _format_key(self, key):
        return "{}:".format(key).ljust(self.KEY_LENGTH - 2 * self._indent)

    def _format_value(self, value):
        return self._format(value, len(str(value))+3)

    def _print_key_value(self, key, value):
        if value is None:
            return
        line = "{indent}{key}{value}"
        print line.format(indent='  ' * self._indent,
                          key=self._format_key(key),
                          value=self._format_value(value if value is not None else 'None'))

    def _print_system_header(self, system):
        self._print_key_value("Name", system.get_name())
        self._print_key_value("Serial", system.get_serial())
        self._print_key_value("Management Address", system.get_management())

    def _print_connections(self, device):
        for connection in device.get_connections():
            self._print_key_value("Connection", "")
            self._indent += 1
            self._print_key_value("Type", connection.get_type())
            self._print_key_value("Initiator", connection.get_initiator())
            self._print_key_value("Target", connection.get_target())
            self._print_key_value("Node", connection.get_node_id())
            self._print_key_value("Port", connection.get_port_id())
            self._print_key_value("Access Path", connection.get_access_path())
            self._print_key_value("Path State", connection.get_state())
            self._indent -= 1

    def _print_device(self, device):
        self._print_key_value("Device", "")
        self._indent += 1
        self._print_key_value("Mount Point", device.get_mount_point())
        self._print_key_value("LUN", device.get_lun())
        self._print_key_value("Type", device.get_type())
        self._print_key_value("SCSI Serial", device.get_scsi_serial())
        self._print_key_value("Volume ID", device.get_volume_id())
        self._print_key_value("Volume Name", device.get_volume_name())
        self._print_key_value("Soft Size", self._size_to_string(device.get_size_in_bytes()))
        self._print_key_value("Multipathed", device.is_multipath())
        self._print_key_value("Multipathing", device.get_multipath_framework_name())
        self._print_key_value("Access Path", device.get_access_path())
        self._print_key_value("Display Name", device.get_display_name())
        self._print_key_value("Connections", "")
        self._indent += 1
        self._print_connections(device)
        self._indent -= 1
        self._indent -= 1
        return

    def _print_system(self, system):
        self._print_key_value("System", '')
        self._indent += 1
        self._print_system_header(system)
        self._print_key_value("Devices", "")
        self._indent += 1
        for device in system.get_devices():
            self._print_device(device)
        self._indent -= 1
        self._indent -= 1


class JSONOutput(OutputStrategy):
    def execute(self, inventory):
        output = [dict(name=system.get_name(),
                       serial=system.get_serial(),
                       management=system.get_management(),
                       devices=[dict(lun=device.get_lun(),
                                     type=device.get_type(),
                                     scsi_serial=device.get_scsi_serial(),
                                     volume_id=device.get_volume_id(),
                                     volume_name=device.get_volume_name(),
                                     size_in_bytes=device.get_size_in_bytes(),
                                     is_multipath=device.is_multipath(),
                                     multipathing_name=device.get_multipath_framework_name(),
                                     mount_point=device.get_mount_point(),
                                     access_path=device.get_access_path(),
                                     display_name=device.get_display_name(),
                                     connections=[dict(type=connection.get_type(),
                                                       initiator=connection.get_initiator(),
                                                       target=connection.get_target(),
                                                       node=connection.get_node_id(),
                                                       port=connection.get_port_id(),
                                                       access_path=connection.get_access_path(),
                                                       state=connection.get_state())
                                                  for connection in device.get_connections()])
                                 for device in system.get_devices()])
                   for system in inventory.get_systems()]
        from json import dumps
        print dumps(output, indent=4)


class SummaryTerminalOutput(OutputStrategy):
    def execute(self, inventory):
        from capacity import byte
        self._inventory = inventory  # pylint: disable=W0201
        number_of_systems = len(self._inventory.get_systems())
        number_of_volumes = self._calculate_number_of_volumes()
        total_size_in_bytes = self._calculate_total_size_in_bytes()
        total_size = 1 * byte
        total_size *= total_size_in_bytes
        message = "INFINIDAT stores {} on top of {} volumes in {} system{plural}."
        formatted_message = message.format(str(total_size).replace('*', ' '),
                                           number_of_volumes, number_of_systems,
                                           plural="" if number_of_systems == 1 else "s")
        print formatted_message

    def _get_volumes_from_system(self, system):
        return filter(lambda device: device.get_type() == "Volume", system.get_devices())

    def _calculate_total_size_in_bytes(self):
        total_size = 0
        for system in self._inventory.get_systems():
            for volume in self._get_volumes_from_system(system):
                total_size += volume.get_size_in_bytes()
        return total_size

    def _calculate_number_of_volumes(self):
        number_of_volumes = 0
        for system in self._inventory.get_systems():
            number_of_volumes += len(self._get_volumes_from_system(system))
        return number_of_volumes


class OutputContext(object):
    def __init__(self, output_stategy_name):
        super(OutputContext, self).__init__()
        self._output_stategy_name = output_stategy_name

    @classmethod
    def get_supported_strategies(cls):
        return dict(brief=BriefTerminalOutput(),
                    detailed=DetailedTerminalOutput(),
                    summary=SummaryTerminalOutput(),
                    json=JSONOutput())

    @classmethod
    def get_preferred_stategy_name(cls):
        return "brief"

    def _get_strategy(self):
        return self.get_supported_strategies()[self._output_stategy_name]

    def execute_strategy(self, inventory, size_units):
        strategy = self._get_strategy()
        strategy.set_size_units(size_units)
        strategy.execute(inventory)


def run(output_stategy_name, size_units='GB', rescan=True):
    from ...rescan.engine import do_rescan
    if rescan:
        sys.stderr.write("\rRescanning... ")
        sys.stderr.flush()
        do_rescan()
    inventory = Inventory()
    sys.stderr.write("\rQuerying devices... ")
    sys.stderr.flush()
    try:
        with progress_bar_context():
            inventory.populate()
    finally:
        sys.stderr.write("\r" + " " * (len("Querying devices... ") + 1) + "\r")
        sys.stderr.flush()
    context = OutputContext(output_stategy_name)
    context.execute_strategy(inventory, size_units)
    if inventory._exc_info:
        raise inventory._exc_info[0], inventory._exc_info[1], inventory._exc_info[2]
