import os
import sys
import platform
from os import makedirs
from unittest.runner import _WritelnDecorator
from logging import getLogger
from infi.vendata.powertools.utils import safe_rescan, round_to_block_size
from infi.vendata.powertools.utils import get_infinidat_multipath_block_devices
from infi.vendata.powertools.utils.tasks import Task, TaskSequence, make_blocking
from infi.vendata.powertools.utils import is_veritas_multipathing_installed
from ...metadata_sync.engine import silent_metadata_sync
from infi.pyutils.lazy import cached_method
from capacity import byte

log = getLogger(__name__)

# pylint: disable=C0103,W0631

class ProvisioningError(Exception):
    pass

def is_letter_mount(string):
    from re import match
    return match("^[A-Za-z]:\\\$", string) is not None


class ProvisioningTask(Task):
    def _is_provisioned_device(self, device):
        return device.get_vendor().get_system_serial() == self.args.box_id and \
               device.get_vendor().get_volume_id() == self.args.volume_id

    def get_storage_device(self):
        block_devices = get_infinidat_multipath_block_devices()
        devices = filter(self._is_provisioned_device, block_devices)
        if devices:
            return devices[0]
        raise ProvisioningError("No matching MPIO device found for volume id {}".format(self.args.volume_id))

    def get_filesystem(self):
        from infi.storagemodel import get_storage_model
        model = get_storage_model()
        filesystems = model.get_mount_manager().get_available_file_systems()
        for filesystem in filesystems:
            if filesystem.get_name() == self.args.filesystem:
                return filesystem

    def rescan(self, extra_predicates=[]):
        from infi.storagemodel import get_storage_model
        from infi.storagemodel.predicates import ScsiDevicesAreReady, MultipathDevicesAreReady, PredicateList
        from infi.vendata.powertools.rescan.engine import clear_unit_attention_from_infinidat_storage_controllers
        from infi.vendata.powertools.rescan.engine import consider_reconfigure_multipathd
        clear_unit_attention_from_infinidat_storage_controllers()
        model = get_storage_model()
        if platform.system() == "Linux":
            # if mapping volumes to LUN0, we'll need to trigger mpio rescan first
            consider_reconfigure_multipathd()
            model.refresh()
        predicates = PredicateList(extra_predicates + [ScsiDevicesAreReady(), MultipathDevicesAreReady()])
        model.rescan_and_wait_for(predicates, timeout_in_seconds=600)
        if platform.system() == "Linux":
            consider_reconfigure_multipathd()
            model.refresh()


class CreateNewVolume(ProvisioningTask):  # pragma: no cover

    def _size(self):
        return str(self.args.volume_size).replace('*', ' ')

    def get_description(self):
        return 'Creating the new {} volume "{}"'.format(self._size(), self.args.volume_name)

    def get_workout(self):
        return ['Create a {volume_type} {volume_size} volume "{volume_name}" of on InfiniBox {box_id}.'.format(
                    volume_name=self.args.volume_name,
                    volume_type=self.args.volume_type,
                    volume_size=self._size(),
                    box_id=self.args.box_id),
                'Map this volume to this host on {}.'.format(
                    'the next available LUN' if self.args.lun is None else 'LUN {}'.format(self.args.lun))
                ]

    @cached_method
    @safe_rescan
    def get_system_object(self):
        """ this method returns a infinisdk.InfiniBox object for the chosen box.
        It gets the management IP address from inquiry
        """
        from ...utils.queries import get_system_object_and_host
        self.system, self.host = get_system_object_and_host(self.args.box_id)

    @cached_method
    def get_pool_object(self):
        self.pool = self.system.pools.get(name=self.args.pool_name)

    @cached_method
    def create_volume_object(self):
        size = round_to_block_size(self.args.volume_size // byte) * byte
        self.volume = self.system.volumes.create(pool=self.pool,
                                                 name=self.args.volume_name,
                                                 size=size,
                                                 provisioning=self.args.volume_type)

    def run(self):
        """:returns: mapping tuple (system, host, volume, lun)"""
        self.get_system_object()
        self.get_pool_object()
        if self.args.lun in self.host.get_luns().luns:
            raise ProvisioningError("Another volume is already mapped to LUN {}".format(self.args.lun))
        self.create_volume_object()
        log.info("Created volume {} inside pool {}".format(self.volume, self.pool))
        self.lun = self.host.map_volume(self.volume, self.args.lun)
        log.info("Mapped volume {} to host {} on lun {}".format(self.volume, self.host, self.lun))
        # the following assignment updates the global 'args' dictionary in the local process. This is important
        # because although this is not really an argument but a runtime thing, it later passes to the other tasks
        # including those being run in worker child processes
        self.args.volume_id = self.volume.get_id()


class InfinidatMultipathDeviceExists(object):
    """A predicate that checks if an Infinidat volume exists"""
    def __init__(self, system_serial, volume_id):
        self.system_serial = system_serial
        self.volume_id = volume_id

    def __call__(self):
        devices_to_query = get_infinidat_multipath_block_devices()
        log.debug("Looking for INFINIDAT volume id {} from system id {}".format(self.volume_id, self.system_serial))
        for device in devices_to_query:
            device.get_scsi_test_unit_ready()
            volume_id = device.get_vendor().get_naa().get_volume_id()
            system_serial = device.get_vendor().get_naa().get_system_serial()
            log.debug("Found INFINIDAT volume id {} from system id {}".format(volume_id, system_serial))
        return any([self.volume_id == device.get_vendor().get_naa().get_volume_id() and
                    self.system_serial == device.get_vendor().get_naa().get_system_serial()
                    for device in devices_to_query])

    def __repr__(self):
        return "<InfinidatMultipathDeviceExists(system_serial={!r}, volume_id={!r})>".format(self.system_serial,
                                                                                             self.volume_id)


class WaitForVolume(ProvisioningTask):  # pragma: no cover
    def get_description(self):
        return "Waiting for volume discovery, this might take a while"

    def run(self):
        predicate = InfinidatMultipathDeviceExists(self.args.box_id, self.args.volume_id)
        self.rescan([predicate])
        self.args.device_display_name = self.get_storage_device().get_display_name()
        # the returned dictionary here goes to the global "args" dict in the parent process so it can use it
        return dict(device_display_name=self.args.device_display_name)


class SolarisLabel(ProvisioningTask):  # pragma: no cover
    def get_description(self):
        return "Labeling disk"

    def get_workout(self):
        return ["Labeling disk with EFI."]

    def run(self):
        from infi.storagemodel import get_storage_model
        from os import path
        if not getattr(self.args, "skip_rescan_before_creating_filesystem", False):
            self.rescan()
        else:
            get_storage_model().refresh()
        storage_device = self.get_storage_device()
        block_access_path = path.basename(storage_device.get_block_access_path()[:-2])  # without s2
        self.make_label(block_access_path)

    def make_label(self, block_access_path):
        raise NotImplementedError


class Solaris10Label(SolarisLabel):
    def make_label(self, block_access_path):
        from pexpect import spawn
        p = spawn("format -D {}".format(block_access_path))
        p.expect(".*selecting {}.*".format(block_access_path))
        p.sendline("fdisk")
        p.expect(".*")
        p.sendline("y")
        p.sendline("label")
        p.expect(".*Ready to label disk, continue?.*")
        p.sendline("y")
        p.expect(".*format.*")
        p.terminate()  # just to make sure


class Solaris11Label(SolarisLabel):
    def make_label(self, block_access_path):
        from infi.vendata.powertools.utils import execute_command
        execute_command(["format", "-L", "efi", "-d", block_access_path])


class SolarisVeritasDiskSetup(ProvisioningTask):  # pragma: no cover
    def get_description(self):
        return "Setting up veritas disk"

    def get_workout(self):
        return ["Setting up veritas disk using vxdisksetup"]

    def run(self):
        from infi.storagemodel import get_storage_model
        from infi.vendata.powertools.utils import execute_command
        from os import path
        if not getattr(self.args, "skip_rescan_before_creating_filesystem", False):
            self.rescan()
        else:
            get_storage_model().refresh()
        storage_device = self.get_storage_device()
        execute_command(["/etc/vx/bin/vxdisksetup", "-if", path.basename(storage_device.get_block_access_path())])  # without s2


class CreateFileSystem(ProvisioningTask):  # pragma: no cover
    def get_description(self):
        return "Creating filesystem"

    def get_workout(self):
        is_an_prefix = any(self.args.filesystem.startswith(char) for char in 'AEIOUaeiou')
        filesystem_prefix = ('an' if is_an_prefix else 'a')
        return ["Create {filesystem_prefix} {filesystem} filesystem on this volume.".format(
            filesystem=self.args.filesystem,
            filesystem_prefix=filesystem_prefix)]

    def run(self):
        from infi.storagemodel import get_storage_model
        from infi.storagemodel.vendor.infinidat.infinibox import ALIGNMENT
        if not getattr(self.args, "skip_rescan_before_creating_filesystem", False):
            self.rescan()
        else:
            get_storage_model().refresh()
        storage_device = self.get_storage_device()
        disk_drive = storage_device.get_disk_drive()
        if os.name == 'nt':
            disk_drive.online() if not disk_drive.is_online() else None
            disk_drive.unset_read_only_attribute() if disk_drive.has_read_only_attribute() else None
        create_partition_table = disk_drive.create_guid_partition_table if hasattr(disk_drive, 'create_guid_partition_table') \
                                 else disk_drive.create_gpt_partition_table
        partition_table = create_partition_table(ALIGNMENT)
        log.info("Created partition table {} for disk {} on device {}".format(partition_table,
                                                                              disk_drive,
                                                                              storage_device))
        filesystem = self.get_filesystem()
        partition = partition_table.create_partition_for_whole_table(filesystem, ALIGNMENT)
        log.info("Created partition {} in partition table {}".format(partition, partition_table))
        if os.name == 'nt':
            partition = self._rescan_and_get_updated_partition()
        filesystem.format(partition)
        silent_metadata_sync()
        log.info("Formatted filesystem {} on partition {}".format(filesystem, partition))

    def _rescan_and_get_updated_partition(self):
        # from jenkins:
        # File "c:\jenkins\workspace\powertools\src\infi\vendata\powertools\volume_provision\engine\__init__.py", line 178, in run
        # filesystem.format(partition)
        # Local variables:
        # 'partition_table': <infi.storagemodel.windows.partition.WindowsGPTPartitionTable object at 0x000000000300DDA0>
        # 'self': <infi.vendata.powertools.volume_provision.engine.CreateFileSystem object at 0x0000000002895E80>
        # 'partition': <infi.storagemodel.windows.partition.WindowsGUIDPartition object at 0x0000000003055198>
        # 'disk_drive': <infi.storagemodel.windows.disk.WindowsDiskDrive object at 0x00000000030074A8>
        # 'filesystem': <infi.storagemodel.windows.filesystem.WindowsFileSystem object at 0x0000000002EFE240>
        # 'storage_device': <MultipathBlockDevice \\?\GLOBALROOT\Device\MPIODisk0 for PHYSICALDRIVE1>
        # 'ALIGNMENT': 65536

        # File "c:\jenkins\workspace\powertools\eggs\infi.storagemodel-0.2.8.post6.g586a370-py2.7.egg\infi\storagemodel\windows\filesystem.py", line 19, in format
        # volume = block_device._get_volume()
        # Local variables:
        # 'block_device': <infi.storagemodel.windows.partition.WindowsGUIDPartition object at 0x0000000003055198>
        # 'self': <infi.storagemodel.windows.filesystem.WindowsFileSystem object at 0x0000000002EFE240>
        # 'args': ()
        # 'kwargs': {}

        # File "c:\jenkins\workspace\powertools\eggs\infi.storagemodel-0.2.8.post6.g586a370-py2.7.egg\infi\storagemodel\windows\partition.py", line 65, in _get_volume
        # raise RescanIsNeeded()
        # Local variables:
        # 'WindowsException': <class 'infi.wioctl.errors.WindowsException'>
        # 'self': <infi.storagemodel.windows.partition.WindowsGUIDPartition object at 0x0000000003055198>

        # RescanIsNeeded
        def get_partition():
            storage_device = self.get_storage_device()
            disk_drive = storage_device.get_disk_drive()
            partition_table = disk_drive.get_partition_table()
            partitions = partition_table.get_partitions()
            return partitions[-1]

        def predicate():
            return get_partition()._get_volume()._path is not None

        self.rescan(extra_predicates=[predicate])
        return get_partition()


class MountVolume(ProvisioningTask):
    def get_description(self):
        return "Mounting filesystem"

    def get_workout(self):
        return ["Mount this volume to {mountpoint}.".format(mountpoint=self.args.mountpoint),
                "Give this volume the label {label}.".format(label=self.args.label)]

    def _create_mount_path(self, mountpoint):
        # no need to create the mount path on Windows, if we are using a drive letter
        if os.name == 'nt' and is_letter_mount(mountpoint):
            return
        if not os.path.exists(mountpoint):
            makedirs(mountpoint)

    def _update_volume_metadata(self, mountpoint):
        from infinisdk.core.exceptions import APICommandFailed
        from infi.vendata.powertools.utils.queries import get_system_object_by_serial
        infinisdk_system = get_system_object_by_serial(self.args.box_id)
        infinisdk_volume = infinisdk_system.volumes.get(id=self.args.volume_id)
        try:
            infinisdk_volume.set_metadata("mount_point", mountpoint)
        except APICommandFailed as error:
            if error.status_code == 403:
                log.warn(str(error))
            else:
                raise

    def run(self):
        from infi.storagemodel import get_storage_model
        from infi.storagemodel.errors import LabelNotSupported
        from infi.pyutils.lazy import clear_cache
        from infi.vendata.powertools.utils import safe_get_device_partitions
        if not getattr(self.args, "skip_rescan_before_mounting", False):
            self.rescan()
        else:
            get_storage_model().refresh()
        partition = safe_get_device_partitions(self.get_storage_device())[0]
        filesystem = self.get_filesystem()
        mountpoint = self.args.mountpoint
        if os.name == 'nt':
            if is_letter_mount(mountpoint + "\\"):
                # this is a shortcut for mounting to a letter on nt
                mountpoint += "\\"
            if not is_letter_mount(mountpoint) and not mountpoint.endswith("\\"):
                # the mount point has to end with "\" for creation in Windows.
                # we add this here and not in diskmanagement because the mount manager also returns the mounts with the
                # trailing slash so we need it to be updated here
                mountpoint += "\\"
                storage_device = self.get_storage_device()
                disk_drive = storage_device.get_disk_drive()
                disk_drive.online() if not disk_drive.is_online() else None
                disk_drive.set_read_only_attribute() if self.args.get("read_only") else disk_drive.unset_read_only_attribute()

        self._create_mount_path(mountpoint)
        block_access_path = partition.get_block_access_path()

        mount_options = {}
        if self.args.get("read_only"):
            if os.name == "nt":
                pass
            elif filesystem.get_name() in ("ext4", ):
                mount_options.update(ro=True, noload=True)
            else:
                mount_options.update(ro=True)

        filesystem.mount(block_access_path, mountpoint, mount_options)
        try:
            # we can only set the label AFTER the mount on Windows, and it still may fail
            filesystem.set_label(block_access_path, self.args.label)
        except LabelNotSupported:
            pass
        self._update_volume_metadata(unicode(mountpoint))
        repository = get_storage_model().get_mount_repository()
        manager = get_storage_model().get_mount_manager()
        silent_metadata_sync()
        clear_cache(manager)
        expected_mountpoint = os.path.abspath(mountpoint.lower())
        for mount_object in manager.get_mounts():
            log.debug("found mountpoint {!r}".format(mount_object.get_mount_point()))
            if mount_object.get_mount_point().lower().rstrip(os.path.sep) == expected_mountpoint.rstrip(os.path.sep):
                repository.add_persistent_mountpoint(mount_object)
                break
        else:
            raise ProvisioningError("Could not find created mount, expected {!r}".format(expected_mountpoint))

def load_tasks(args):  # pragma: no cover
    from infi.os_info import get_platform_string
    tasks = [CreateNewVolume(args), make_blocking(WaitForVolume(args))]
    args.skip_rescan_before_creating_filesystem = True
    args.skip_rescan_before_mounting = True
    if args.do_filesystem:
        tasks.append(make_blocking(CreateFileSystem(args)))
        if args.do_mount:
            tasks.append(make_blocking(MountVolume(args)))
    elif 'solaris' in get_platform_string():
        if is_veritas_multipathing_installed():
            tasks.append(make_blocking(SolarisVeritasDiskSetup(args)))
        else:
            if '10' == get_platform_string().split('-')[1]:
                tasks.append(Solaris10Label(args))
            else:
                tasks.append(make_blocking(Solaris11Label(args)))
    return TaskSequence(tasks)

def get_workout(arguments):
    args = cli_arguments_to_namespace(arguments)
    msg = "This tool is going to:\n" + load_tasks(args).format_workout()
    return msg

def get_end_result(args):  # pragma: no cover
    msg = "All done.\n"
    msg += "The volume {} is represented by {}.".format(args.volume_name,
                                                        args.device_display_name)
    if args.do_filesystem and args.do_mount:
        msg = msg[:-1] + ", and it is mounted to {}.".format(args.mountpoint)
    return msg

def name_to_label(name):
    return name[:32 if os.name == 'nt' else 16]

def cli_arguments_to_namespace(arguments):
    from munch import Munch
    return Munch(volume_type="thin" if arguments["--thin"].lower() in ("yes",) else "thick",
                 volume_name=arguments["--name"],
                 lun=arguments.get("--lun"),
                 volume_size=arguments["<size>"],
                 mountpoint=arguments["--mount"],
                 pool_name=arguments["--pool"],
                 filesystem=arguments["--filesystem"],
                 box_id=int(arguments["--system"]),
                 label=name_to_label(arguments["--name"]),
                 do_mount=not arguments["--no-mount"],
                 do_filesystem=not arguments["--no-filesystem"])

def provisioning(arguments, stream=_WritelnDecorator(sys.stdout)):  # pragma: no cover
    from infi.storagemodel import get_storage_model
    args = cli_arguments_to_namespace(arguments)
    load_tasks(args).run(stream)
    stream.writeln("="*79)
    stream.flush()
    print get_end_result(args)
    get_storage_model().refresh()
    return args

def validate_create_filesystem(device, filesystem_type):
    from infi.vendata.powertools.utils import has_filesystem
    from infi.vendata.powertools.cli.validate import validate_filesystem_size
    from infi.vendata.powertools.utils.queries import is_device_read_only
    if has_filesystem(device):
        raise ProvisioningError("Filesystem already exists")
    if is_device_read_only(device):
        raise ProvisioningError("Device is read only")
    device_size = device.get_size_in_bytes() * byte
    validation_result = validate_filesystem_size(filesystem_type, device_size)
    if validation_result:
        raise ProvisioningError(validation_result)

def create_filesystem(device, filesystem_type, silent=False):
    from munch import Munch
    from infi.storagemodel import get_storage_model
    validate_create_filesystem(device, filesystem_type)
    args = Munch()
    args.filesystem = filesystem_type
    args.volume_id = device.get_vendor().get_volume_id()
    args.box_id = device.get_vendor().get_system_serial()
    operation = make_blocking(CreateFileSystem(args))
    if silent:
        TaskSequence([operation]).step()
    else:
        TaskSequence([operation]).run()
    get_storage_model().refresh()

def mount(device, mount_point, read_only=False, silent=False):
    from munch import Munch
    from infi.storagemodel import get_storage_model
    from infi.vendata.powertools.utils import is_mounted, safe_get_device_partitions
    from infi.vendata.powertools.utils.queries import is_device_read_only
    if is_mounted(device):
        raise ProvisioningError("Volume is already mounted")
    if os.name == 'nt' and is_device_read_only(device):
        raise ProvisioningError("Read only volumes can not be mounted on Windows system")
    args = Munch()
    args.mountpoint = mount_point
    args.label = name_to_label(device.get_vendor().get_volume_name())
    args.volume_id = device.get_vendor().get_volume_id()
    args.box_id = device.get_vendor().get_system_serial()
    args.read_only = read_only
    args.skip_rescan_before_mounting = True  # volume already mapped, no need to rescan
    partitions = safe_get_device_partitions(device)
    if not partitions:
        raise ProvisioningError("There are no partitions in the volume")
    partition = partitions[0]
    args.filesystem = partition.get_current_filesystem().get_name()
    # HIP-1433
    # lets check now that the filesystem we see on the partition can be mounted,
    # instead of doing this check in the blocking child process later,
    available_filesystems = set([item.get_name() for item in get_storage_model().get_mount_manager().get_available_file_systems()])
    if args.filesystem not in available_filesystems:
        msg = "Invalid/unsupported filesystem ({}) found on partition {}"
        raise ProvisioningError(msg.format(args.filesystem, partition.get_block_access_path()))
    operation = make_blocking(MountVolume(args))
    if silent:
        TaskSequence([operation]).step()
    else:
        TaskSequence([operation]).run()
    get_storage_model().refresh()

def in_provisioning(arguments):
    from infinisdk.core.exceptions import APICommandFailed
    try:
        provisioning(arguments)
        return 0
    except APICommandFailed as exc:
        try:
            message = exc.message
            log.error("Provisioning Infinipy Error: {}".format(message))
            print "ERROR: {}".format(message)
            return 3
        except:
            pass
        # couldn't parse the infinisdk error message to display somehting sensible to the user
        # reraise original infinisdk traceback and let debug_program catch this as a generic error
        raise
    except ProvisioningError as error:
        log.error("Provisioning Error: {}".format(error.message))
        print "ERROR: {}".format(error.message)
        return 2
