import sys
import multiprocessing
import threading
import capacity
from infi.storagemodel.connectivity import FCConnectivity
from infi.os_info import get_platform_string
from infi.pyutils.lazy import cached_method
from multiprocessing.queues import SimpleQueue
import os

class SpeedTestException(Exception):
    pass

def perform_read(device_path, block_size, total_disk_size, ipc):
    if 'solaris-11' in get_platform_string() and device_path.endswith('s2'):
        device_path = device_path[:-2]
    with open(device_path, "rb") as fd:
        bytes_read = 0
        ipc.thread_ready_queue.put("Thread ready")
        ipc.start_event.wait()
        while not ipc.close_event.is_set():
            read_buf = fd.read(block_size)
            ipc.io_bytes_queue.put(len(read_buf))
            bytes_read += len(read_buf)
            if bytes_read >= total_disk_size:
                fd.seek(0)
                bytes_read = 0

def thread_func(device_path, block_size, total_disk_size, ipc):
    perform_read(device_path, block_size, total_disk_size, ipc)

def process_func(thread_count, device_path, block_size, total_disk_size, ipc):
    threads = []
    for i in range(thread_count):
        thread = threading.Thread(target=thread_func, args=(device_path, block_size, total_disk_size, ipc))
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()

def get_storagemodel_device(device_path):
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
    from infi.storagemodel.base.multipath import MultipathBlockDevice
    block_devices = get_infinidat_block_devices()
    for block_device in block_devices:
        if block_device.get_block_access_path() == device_path:
            if not isinstance(block_device, MultipathBlockDevice):
                msg = "{} is not a multipath device"
                raise SpeedTestException(msg.format(device_path))
            return block_device
    raise SpeedTestException("Couldn't find matching device for {}".format(device_path))

def data_collector_func(device_path, process_count, ipc):
    device = get_storagemodel_device(device_path)
    bytes_read = 0
    io_count = 0
    ipc.start_event.wait()
    last_path_statistics = [path.get_io_statistics() for path in device.get_paths()]
    # on Unix systems, the reads are always cached so we get lower results in the path statistics. We need
    # to multiply the number of IOs by the number of processes (all the processes make the same IOs)
    multiplier = 1 if os.name == 'nt' else process_count
    avg_bytes = 0
    timer_count = 0
    while True:
        if ipc.io_bytes_queue._reader.poll(0.1):
            bytes_read += ipc.io_bytes_queue.get()
            io_count += 1
        if ipc.close_event.is_set() and ipc.io_bytes_queue.empty():
            break
        if ipc.timer_event.is_set():
            path_statistics = [path.get_io_statistics() for path in device.get_paths()]
            # add results to per second queue
            ipc.result_queue.put((bytes_read, io_count,
                                 [((path_stats.bytes_read - last_path_stats.bytes_read) * multiplier,
                                   (path_stats.read_io_count - last_path_stats.read_io_count))
                                    for last_path_stats, path_stats in zip(last_path_statistics, path_statistics)]))
            # update avg_bytes and send to verbose queue
            if ipc.verbose_queue is not None:
                timer_count += 1
                avg_bytes = ((avg_bytes * (timer_count - 1)) + bytes_read) / timer_count
                ipc.verbose_queue.put(avg_bytes)
            # reset
            bytes_read = 0
            io_count = 0
            last_path_statistics = path_statistics
            ipc.timer_event.clear()

def Gbps_to_Bps(Gbps):
    # gigabit per second to bytes per second
    #             Mbps   Kbps   bps   Bps
    return Gbps * 1024 * 1024 * 1024 / 8

def format_bytes(bytes, unit=None):
    if unit is None:
        return str(capacity.byte * bytes)
    return str(capacity.byte * bytes / unit) + str(unit)[1:]

def queue_to_iterator(queue):
    while not queue.empty():
        yield queue.get()

class PerformanceEngine(object):
    def __init__(self,
                 device_path,
                 process_count=10,
                 per_process_thread_count=1,
                 block_size=1024 * 1024 / 2,
                 time_to_run=60,
                 print_to_verbose_queue=False):
        self.device_path = device_path
        self.process_count = process_count
        self.per_process_thread_count = per_process_thread_count
        self.block_size = block_size
        self.time_to_run = time_to_run
        self._print_to_verbose_queue = print_to_verbose_queue
        self._create_ipc_objects()

    def _timer(self):
        self._log("Running for {} seconds...".format(self.time_to_run))
        for i in range(self.time_to_run):
            if self.ipc.close_event.wait(timeout=1):
                # normally this function will set close_event, but cancel may also be called
                break
            self.ipc.timer_event.set()
            if self._print_to_verbose_queue:
                self.ipc.verbose_queue.put("Running for {} seconds...\n{} Remaining".format(self.time_to_run,
                                                                                        self.time_to_run - i - 1))
            else:
                # sys.stdout.flush will get stuck in the GUI (we don't have a console window there). So write to
                # stdout here only if not using the GUI, which is when we don't print to verbose queue
                sys.stdout.write(".")
                sys.stdout.flush()
        self.ipc.close_event.set()

    def _get_port_speeds(self):
        from infi.hbaapi import get_ports_collection
        initiator_ports = get_ports_collection().get_ports()
        wwn_to_speed = dict((port.port_wwn, port.port_speed) for port in initiator_ports)
        return wwn_to_speed

    @cached_method
    def _get_wire_speed(self):
        """ Returns max. wire speed in bytes per second """
        paths = self.device.get_paths()
        connectivities = [path.get_connectivity() for path in paths
                          if isinstance(path.get_connectivity(), FCConnectivity)]
        initiator_speed = sum(self._get_port_speeds().values())
        # we can't know the speed of the target ports, but we assume it's 8Gbps
        # because it's what infiniboxes have
        target_wwns = set(connectivity.get_target_wwn() for connectivity in connectivities)
        target_speed = len(target_wwns) * 8
        return Gbps_to_Bps(min(initiator_speed, target_speed))

    @cached_method
    def _get_target_info(self, target_wwn):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
        from infi.storagemodel.base.multipath import MultipathDevice
        for storage_controller in get_infinidat_block_devices_and_controllers__mapped_to_lun0():

            if isinstance(storage_controller, MultipathDevice):
                for path in storage_controller.get_paths():
                    if path.get_connectivity().get_target_wwn() == target_wwn:
                        # we can't get rptg per path, so we use this
                        return str(target_wwn)[-2], str(target_wwn)[-1]

            elif storage_controller.get_connectivity().get_target_wwn() == target_wwn:
                port = storage_controller.get_vendor().get_fc_port()
                return port.get_node_id(), port.get_port_id()

    def _print_path_table(self):
        if self._print_to_verbose_queue:
            # no prints in GUI mode
            return
        print
        print "Device paths:"
        path_row_format = "{:10s}{:10s}{:18s}{:18s}{:8s}{:8s}"
        if "solaris" in get_platform_string() or "aix" in get_platform_string():
            # on AIX and Solaris, the "target" of the HCTL is fake and longer
            path_row_format = "{:10s}{:27s}{:18s}{:18s}{:8s}{:8s}"
        print path_row_format.format("Path #",
                                     "HCTL",
                                     "Initiator WWN",
                                     "Target WWN",
                                     "NodeID",
                                     "PortID")
        for i, path in enumerate(self.device.get_paths()):
            connectivity = path.get_connectivity()
            if not isinstance(connectivity, FCConnectivity):
                continue
            target_node_id, target_port_id = self._get_target_info(connectivity.get_target_wwn())
            print path_row_format.format("Path #{}".format(i+1),
                                         str(path.get_hctl()),
                                         str(connectivity.get_initiator_wwn()),
                                         str(connectivity.get_target_wwn()),
                                         str(target_node_id),
                                         str(target_port_id))
        print

    def _check_path_balance(self, path_performances):
        # the difference between the throughput of each path and the average throughput of all the paths should
        # not exceed 10% of the average
        path_bytes = (total_path_bytes for total_path_bytes, total_path_io_count in path_performances)
        average_path_bytes = sum(path_bytes) / len(path_performances)
        if average_path_bytes == 0:
            # failed to read path statistics or something like that...
            return False
        for total_path_bytes, total_path_io_count in path_performances:
            if abs(average_path_bytes - total_path_bytes) / float(average_path_bytes) > 0.1:
                return False
        return True

    def _print_result(self):
        per_second_results = list(queue_to_iterator(self.ipc.result_queue))
        if len(per_second_results) == 0:
            return

        total_bytes = 0
        total_io_count = 0
        path_performances = []
        for i in range(len(self.device.get_paths())):
            path_performances.append([0, 0])
        for bytes_read_in_second, io_in_second, path_performance_in_second in per_second_results:
            total_bytes += bytes_read_in_second
            total_io_count += io_in_second
            for i, (path_bytes, path_io) in enumerate(path_performance_in_second):
                path_performances[i][0] += path_bytes
                path_performances[i][1] += path_io
        avg_io_count = total_io_count / len(per_second_results)
        avg_bytes = total_bytes / len(per_second_results)

        wire_speed = self._get_wire_speed()
        if wire_speed == 0:
            throughput_str = "Could not determine wire speed."
        else:
            percent_of_wire_speed = float(avg_bytes) / float(wire_speed / 10.0 * 8) * 100
            wire_speed_str = format_bytes(wire_speed * 8).replace("B", "b")     # show in bits, not bytes
            throughput_str = "Average throughput is {:.2f}% of wire speed ({})".format(percent_of_wire_speed,
                                                                                   wire_speed_str)
        are_path_balanced = self._check_path_balance(path_performances)
        path_str = "Paths are {}balanced.".format("" if are_path_balanced else "not ")

        if self._print_to_verbose_queue:
            self.ipc.verbose_queue.put("Result:\n{}\n{}".format(throughput_str, path_str))
            # no prints in GUI mode
            return

        print
        print "{:10s}{:20s}{:13s}".format("",
                                          "Average Throughput",
                                          "Average IOPS ({})".format(format_bytes(self.block_size)))
        print "{:10s}{:20s}{:13s}".format("Total", format_bytes(avg_bytes, capacity.MB) + "/s", str(avg_io_count))
        for i, (total_path_bytes, total_path_io_count) in enumerate(path_performances):
            avg_path_bytes = total_path_bytes / len(per_second_results)
            avg_path_bytes = avg_path_bytes
            avg_path_io_count = total_path_io_count / len(per_second_results)
            print "{:10s}{:20s}{:13s}".format("Path #{}".format(i+1),
                                              format_bytes(avg_path_bytes, capacity.MB) + "/s",
                                              str(avg_path_io_count))
        print
        print throughput_str
        print path_str

    def _create_io_processes(self):
        total_disk_size = self.device.get_size_in_bytes()
        # remove twice the block size, because we can't read to the very end of the disk, plus we don't want to
        # read partial blocks at the end
        total_disk_size -= self.block_size * 2

        self.processes = []
        # start all IO-performing processes
        for i in range(self.process_count):
            process = multiprocessing.Process(target=process_func,
                                              args=(self.per_process_thread_count, self.device_path, self.block_size, total_disk_size, self.ipc))
            process.start()
            self.processes.append(process)

    def _manage_processes(self):
        """ Actual performance engine: after all IO threads are up, manage them """

        # wait for all threads to load and then start
        # (each thread writes to thread_ready_queue and then waits for start_event)
        for i in range(self.process_count * self.per_process_thread_count):
            self.ipc.thread_ready_queue.get()

        self.ipc.start_event.set()

        # signal data collector thread every second, and then signal close event when time is up
        self._timer()
        self._join_children()

    def _join_children(self):
        # wait for all of our subprocesses to stop
        for process in self.processes:
            process.join(timeout=5)

        self.data_collector_thread.join(timeout=5)

    def get_verbose_queue(self):
        return self.ipc.verbose_queue

    def _create_ipc_objects(self):
        from munch import Munch
        self.ipc = Munch()
        self.ipc.thread_ready_queue = SimpleQueue()         # used by threads to signal they are alive
        self.ipc.io_bytes_queue = SimpleQueue()             # contains number of IOs in bytes performed by threads
        self.ipc.result_queue = SimpleQueue()               # contains per-second result of total IO and path statistics
        self.ipc.start_event = multiprocessing.Event()      # signals all the threads to start after they are all up
        self.ipc.close_event = multiprocessing.Event()
        self.ipc.timer_event = multiprocessing.Event()      # signaled every second to output data to result_queue
        self.ipc.verbose_queue = None
        if self._print_to_verbose_queue:
            self.ipc.verbose_queue = SimpleQueue()          # contains all average throughput values through the test

    def _setup_children(self):
        self.data_collector_thread = multiprocessing.Process(target=data_collector_func,
                                                             args=(self.device_path, self.process_count, self.ipc))
        self.data_collector_thread.start()
        self._create_io_processes()

    def _log(self, msg):
        if self._print_to_verbose_queue:
            self.ipc.verbose_queue.put(msg)
        else:
            print msg

    def cancel(self):
        self.ipc.close_event.set()

    def run(self):
        self.device = get_storagemodel_device(self.device_path)

        if self._print_to_verbose_queue:
            self.ipc.verbose_queue.put(self._get_wire_speed() / 10.0 * 8.0)
        else:
            print "Running performance test on {}".format(self.device_path)

        self._print_path_table()

        self._log("Setting up...")
        self._setup_children()
        self._manage_processes()         # actual engine
        self._print_result()


def _cleanup():
    import gevent
    # this is a partial copy of gipc.gipc._child

    # After fork on POSIX-compliant systems, gevent's state is inherited by the
    # child which may lead to undesired behavior, such as greenlets running in
    # both, the parent and the child. Therefore, on Unix, gevent's and libev's
    # state is reset before running the user-given function.
    if os.name == 'nt':
        return
    # `gevent.reinit` calls `libev.ev_loop_fork()`, which reinitialises
    # the kernel state for backends that have one. Must be called in the
    # child before using further libev API.
    gevent.reinit()
    hub = gevent.get_hub()
    # Delete threadpool before hub destruction, otherwise `hub.destroy()`
    # might block forever upon `ThreadPool.kill()` as of gevent 1.0rc2.
    del hub.threadpool
    hub._threadpool = None
    # Destroy default event loop via `libev.ev_loop_destroy()` and delete
    # hub. This orphans all registered events and greenlets that have been
    # duplicated from the parent via fork().
    hub.destroy(destroy_loop=True)


class MultiprocessingPerformanceEngine(PerformanceEngine):
    def run(self):
        # HIP-1164
        # infinihost-gui, the multiprocessing parent process has greenlets running,
        # which we don't want in the forked child process
        # so we need to clean them up like gipc.gipc, but unlike gipc.gipc we don't want to create a new hub
        # so we use a similar version of gipc.gipc._clenaup
        _cleanup()
        super(MultiprocessingPerformanceEngine, self).run()
